
""" Defintion of all the W24Measure class its support structures


Author: Jochen Mattes - Werk24
"""
from typing import List, Tuple

from pydantic import BaseModel, UUID4


class W24LeaderLabel(BaseModel):
    blurb: str
    """ Uninterpreted string representation of the leader
    """


class W24Leader(BaseModel):
    """ Leaders indicate which location
    """

    leader_id: UUID4
    """ Unique id of the leader
    """

    anchor_coordinates: List[Tuple[float, float]]
    """
    List of x,y tuples indicating the coordinates that
    are referenced by the leader. Be aware that leaders
    can be associated with a multitude of points.
    The x,y coordinates are normalized by the widht/height
    of the sectional image
    """

    text_coordinate: Tuple[float, float]
    """ Tuple indicating the bottom left corner of the
    leader's text box. The value is normalized by the width/heights
    of the sectional image
    """

    label: W24LeaderLabel
    """ Leader label
    """

    associated_measure_ids: List[UUID4] = []
    """ UUID List of measures that were derived from the
    current leader. Refer to W24Measure
    """



""" Websocket-part of the Werk24 client
"""
import json
from types import TracebackType
from typing import Optional, Type, AsyncGenerator

import websockets
from pydantic import ValidationError
from websockets.client import WebSocketClientProtocol
from werk24.exceptions import ServerException, UnauthorizedException
from werk24.models.techread import W24TechreadCommand, W24TechreadMessage

from .auth_client import AuthClient


class TechreadClientWss:
    """ TechreadClient subpart that handles the websocket
    communication with the server.
    """

    def __init__(self, techread_server_wss: str, techread_version: str):
        self._auth_client: Optional[AuthClient] = None
        self._techread_server_wss = techread_server_wss
        self._techread_version = techread_version
        self._techread_session_wss: Optional[WebSocketClientProtocol] = None

    async def __aenter__(
            self
    ) -> 'TechreadClientWss':
        """ Enter the session with the wss server

        Raises:
            RuntimeError  -- Raise when the developer enters the session
                without having called register_auth_client()

        Returns:
            TechreadClientWss -- instance with activated session
        """

        # make sure that we have an AuthClient
        if self._auth_client is None:
            raise RuntimeError(
                "You need to call register_auth_client() before you can start"
                + " the session")

        # make the endpoint
        endpoint = "wss://{}/{}".format(
            self._techread_server_wss,
            self._techread_version)

        # make the ehaders
        headers = [("Authorization", f"Bearer {self._auth_client.token}")]

        # now make the session
        self._techread_session_wss = await websockets.connect(
            endpoint,
            extra_headers=headers)

        # return ourselfves
        return self

    async def __aexit__(
            self,
            exc_type: Optional[Type[BaseException]],
            exc_value: Optional[BaseException],
            traceback: Optional[TracebackType]
    ) -> None:

        """ Close the session
        """
        if self._techread_session_wss is not None:
            await self._techread_session_wss.close()

    def register_auth_client(self, auth_client: AuthClient) -> None:
        """Register the reference to the authentication service

        Arguments:
            auth_client {AuthClient} -- Reference to Authentication
                client
        """
        self._auth_client = auth_client

    async def send_command(
            self,
            action: str,
            message: str = "{}"
    ) -> None:
        """ Send a command to the websocket.

        The function wrapps your action and message into
        a W24TechreadCommand object, translates it to
        json and sends it to the server.

        Be sure to collect the server response from the
        socket using recv_message()

        Arguments:
            action {str} -- Action that is requested
            message {str} -- Auxilliary data that you wnat to send along
                with the message. To keep it easily expandable, we use
                a json encoded string.

        Raises:
            RuntimeError  -- Raise when the developer tries to send a command
                without entering the profile
        """

        # make sure that we have an AuthClient
        if self._techread_session_wss is None:
            raise RuntimeError(
                "You need to call enter the profile before sending a command")

        # make the command
        command = W24TechreadCommand(action=action, message=message)

        # send the the command
        await self._techread_session_wss.send(command.json())

    async def recv_message(self) -> W24TechreadMessage:
        """ Receive a message from the websocket and interpret
        the result as W24TechreadMessage

        Raises:
            RuntimeError  -- Raise when the developer tries to send a command
                without entering the profile

        Returns:
            W24TechreadMessage -- interpreted message
        """

        # make sure that we have an AuthClient
        if self._techread_session_wss is None:
            raise RuntimeError(
                "You need to call enter the profile before receiving command")

        # wait for the websocket to say something
        message_raw = str(await self._techread_session_wss.recv())

        # process the message
        message = await self._process_message(message_raw)

        # and finally return
        return message

    @staticmethod
    async def _process_message(message_raw: str) -> W24TechreadMessage:
        """ Interpret the raw websocket message and
        turn it inot a W24TechreadMessage

        Arguments:
            message_raw {str} -- Raw message

        Raises:
            UnauthorizedException: Exception is raised
                when you requested an action that you
                have no priviledges for (or that does
                not exist)

            ServerException: Exception is raised when
                the server did not respond as expected

        Returns:
            W24TeachreadMessage -- interpreted message
        """

        # interpret and return
        try:
            message = W24TechreadMessage.parse_raw(message_raw)

        # if that failes, we are probably receiving a
        # message from the gateway directly
        except ValidationError:

            # The Gateway responds with the format
            # {"message": str, "connectionId":str, "requestId":str}
            # Obtain the message
            response = json.loads(message_raw)
            message = response.get('message')

            # raise a specific exception if the
            # requested action was forbidden
            if message == 'Forbidden':
                raise UnauthorizedException("Requested Action forbidden")

            # otherwise fail with an UnknownException
            raise ServerException(
                f"Unexpected server response '{message_raw}'.")

        return message

    async def listen(self) -> AsyncGenerator[W24TechreadMessage, None]:
        """ Simple generator that waits for
        messages on the websocket, interprets
        them and yields them

        Yields:
            W24TechreadMessage -- interpreted message from the socket

        Raises:
            RuntimeError  -- Raise when the developer tries to send a command
                without entering the profile
        """

        # make sure that we have an AuthClient
        if self._techread_session_wss is None:
            raise RuntimeError(
                "You need to call enter the profile before listening")

        # wait for incoming messages
        async for message_raw in self._techread_session_wss:

            # process the message and return them to the caller
            yield await self._process_message(str(message_raw))



import argparse

from werk24.cli import utils
from werk24.techread_client import W24TechreadClient


async def main(
    args: argparse.Namespace
) -> None:
    """
    Interact with the Authentication Client through the API

    Args:
        args(argparse.Namespace): CLI args generated by
        argparse
    """

    # get the client instance and handle
    # potential errors
    client = utils.make_client()

    # then check whether we can do anything for the
    # caller
    if args.ask_jwt_token:
        await _handle_ask_jwt_token(client)

async def _handle_ask_jwt_token(
    client: W24TechreadClient,
) -> None:
    """ Steal a valid JWT token from the authentication
    client and print it to the console

    Args:
        client (W24TechreadClient): Current client
    """

    # start a session with the client to obtain a valid
    # token
    async with client as session:
        token =session._auth_client.token
        print(f"Received Token: {token}")


import abc
from decimal import Decimal
from enum import Enum
from typing import Any, Dict, Optional

from pydantic import BaseModel

from .general_tolerances import W24GeneralTolerancesStandard


class W24SizeToleranceType(str, Enum):
    """ Enum of the supported tolerations
    """
    APPROXIMATION = "APPROXIMATION"
    FIT_SIZE_ISO = "FIT_SIZE_ISO"
    GENERAL_TOLERANCES = "GENERAL_TOLERANCES"
    MINIMUM = "MINIMUM"
    MAXIMUM = "MAXIMUM"
    OFF_SIZE = "OFF_SIZE"
    THEORETICALLY_EXACT = "THEORETICALLY_EXACT"
    REFERENCE = "REFERENCE"


class W24SizeTolerance(BaseModel, abc.ABC):
    """ Abstract Base Class to cover the Tolerations

    Attributes:
        blurb (str): String representation for human consumption

        toleration_type (W24SizeToleranceType):  Toleration Type for
            deserialization
    """
    toleration_type: W24SizeToleranceType

    blurb: str


class W24ToleranceGradeWarning(str, Enum):
    """ Warnings associated with the Tolerance Grade.
    """

    SIZE_LARGER_THAN_NORM = "SIZE_LARGER_THAN_NORM"
    TOLERANCE_WIDTH_SMALLER_THAN_NORM = "TOLERANCE_WIDTH_SMALLER_THAN_NORM"
    TOLERANCE_WIDTH_LARGER_THAN_NORM = "TOLERANCE_WIDTH_LARGER_THAN_NORM"


class W24ToleranceGrade(BaseModel):
    """ Tolerance Grade following ISO 286-1

    Attributes:
        blurb (Optional[str]): Blurb for human consumption.

        grade (Optional[str]): Tolerance Grade in the range
            of IT01 to IT18. None if the tolerance is outside the
            normed region

        warning (Optional[W24ToleranceGradeWarning]): the norm is
            limited to the size range (0 - 3150mm) and the
            tolerance range of IT1 - IT18. When we reach the
            ends of this norm, we return a warning.

        NOTE: when a tolerance is outside the norm, we still
            return the closest matching grade.

        NOTE: when tolerances are implausible (e.g., 3+/-6),
            Werk24 will strip the tolerance completely and
            return an untolerated measure of nominal size 3.

    """
    blurb: str

    grade: Optional[str]

    warning: Optional[W24ToleranceGradeWarning]


class W24SizeToleranceFitsizeISO(W24SizeTolerance):
    """ ISO fit size tolerations

    Attributes:
        blurb (str): Text representation for human consumption.

        deviation_lower (Decimal): Lower deviation from the
            nominal size

        deviation_upper (Decimal): Upper deviation from the
            nominal size

        fundamental_deviation (str): Fundamental deviation of
            the fit (e.g., 'H' for a 'H7' fit)

        tolerance_grade (Optional[int]): Tolerance Grade corresponding
            to ISO 286-1. In German IT-Grad.
    """
    toleration_type = W24SizeToleranceType.FIT_SIZE_ISO

    blurb: str

    deviation_lower: Decimal

    deviation_upper: Decimal

    fundamental_deviation: str

    tolerance_grade: W24ToleranceGrade

    @property
    def deviation_width(self) -> Decimal:
        """ Deviation Width

        Returns:
            Decimal: Deviation width
        """
        return self.deviation_upper-self.deviation_lower


class W24SizeToleranceReference(W24SizeTolerance):
    """ Measures written in brackets are Reference
    Measures that are not tolerated.

    Attributes:
        tolerantion_type (W24SizeToleranceType): Reference
            Tolerance Type

        blurb (str): Text representation for human consumption.
    """
    toleration_type = W24SizeToleranceType.REFERENCE


class W24SizeToleranceOffSize(W24SizeTolerance):
    """ Off-size based tolerances

    Attributes:
        blurb (str): Text representation for human consumption.

        deviation_lower (Decimal): Lower deviation from the
            nominal size

        deviation_upper (Decimal): Upper deviation from the
            nominal size

        tolerance_grade (int): Tolerance Grade corresponding to
            ISO 286-1. In German IT-Grad.
    """
    toleration_type = W24SizeToleranceType.OFF_SIZE

    deviation_lower: Decimal

    deviation_upper: Decimal

    tolerance_grade: W24ToleranceGrade


class W24SizeToleranceGeneral(W24SizeTolerance):
    """ General Tolerances

    Attributes:
        tolerance_type: W24SizeToleranceType General Tolerances
            as key to differentiate between the different
            tolerance types

        blurb: Blurb of the General Tolerance following the
            pattern of W24ToleranceOffSize

        standard: GeneralToleranceStandard that is specified in
            the TitleBlock. If the TitleBlock is not requested
            of it no GeneralTolerance was found, this is set to
            None.

        standard_class: GeneralTolerance Class if known. None
            otherwise.

        deviation_lower: Lower deviation from the nominal size
            if the General Tolerance is known. None otherwise.

        deviation_upper: Upper deviation from the nominal size
            if the General Tolerance is known. None otherwise.

        tolerance_grade (int): Tolerance Grade corresponding to
            ISO 286-1. (German: IT-Grad).
    """
    toleration_type = W24SizeToleranceType.GENERAL_TOLERANCES

    blurb: str = ""

    standard: Optional[W24GeneralTolerancesStandard]

    standard_class: Optional[str]

    deviation_lower: Optional[Decimal]

    deviation_upper: Optional[Decimal]

    tolerance_grade: Optional[W24ToleranceGrade]


class W24SizeToleranceTheoreticallyExact(W24SizeTolerance):
    """ Theoretically Exact Measures after ISO 5458
    must not be tolerated. They are indicated by a small
    rectangular frame.

    Example:
        +------+
        |  15  |
        +------+

    !!! note
        In practice, we see Technical drawings contain
        tolerated, theoretically exact measures.
        Example:
            +------------+
            | 15 +/- 0.2 |
            +------------+
        In these situations the toleration takes priority.
    """
    toleration_type = W24SizeToleranceType.THEORETICALLY_EXACT


class W24SizeToleranceMinimum(W24SizeTolerance):
    """ Minimum Size of a measure
    Example:
        min. 15
    """
    toleration_type = W24SizeToleranceType.MINIMUM


class W24SizeToleranceMaximum(W24SizeTolerance):
    """ Maximum Size of a measure
    Example:
        max 15
    """
    toleration_type = W24SizeToleranceType.MAXIMUM


class W24SizeToleranceApproximation(W24SizeTolerance):
    """ Approximation of a measure
    Example:
        ~ 15
        ca. 14
    """
    toleration_type = W24SizeToleranceType.APPROXIMATION


class W24SizeType(str, Enum):
    """ Enum describing the different size types
    """
    NOMINAL = "NOMINAL"
    DIAMETER = "DIAMETER"
    WIDTH_ACROSS_FLATS = "WIDTHS_ACROSS_FLATS"


class W24Size(BaseModel, abc.ABC):
    """ Abstract Base Class for the Sizes

    Attributes:
        blurb: Blurb for human consumption

        size_type: Size type for deserialization

        nominal_size: Unitless nominal size. The unit
            is it attached to the parent object, that
            also defines the toleration.
    """
    blurb: str

    size_type: W24SizeType

    nominal_size: Decimal


class W24SizeNominal(W24Size):
    """ Exactly your nominal size
    """
    size_type = W24SizeType.NOMINAL


class W24SizeDiameter(W24Size):
    """ Diameter size
    """
    size_type = W24SizeType.DIAMETER


class W24SizeWidthsAcrossFlats(W24Size):
    """ Width across flats / Wrench Sizes

    Attributes:
        width_across_flats: Size across flats
            aka. wrench size.
    """
    size_type = W24SizeType.WIDTH_ACROSS_FLATS

    width_across_flats: Decimal


def parse_tolerance(
    raw: Dict[str, Any]
) -> Optional[W24SizeTolerance]:
    """ Pydantic does not automatically return the correct
    W24SizeTolerance object. This function looks at the toleration_type
    attribute and returns the correct W24SizeTolerance subclass

    Args:
        size_tolerance_raw (Dict[str, str]): Raw Dictionary of
            the size tolerance

    Returns:
        W24SizeTolerance: Correctly deserialized Size Tolerance
    """
    # get the class in question
    type_ = raw.toleration_type \
        if hasattr(raw, 'toleration_type')\
        else raw.get('toleration_type')

    class_ = {
        W24SizeToleranceType.APPROXIMATION: W24SizeToleranceApproximation,
        W24SizeToleranceType.FIT_SIZE_ISO: W24SizeToleranceFitsizeISO,
        W24SizeToleranceType.GENERAL_TOLERANCES: W24SizeToleranceGeneral,
        W24SizeToleranceType.MINIMUM: W24SizeToleranceMinimum,
        W24SizeToleranceType.MAXIMUM: W24SizeToleranceMaximum,
        W24SizeToleranceType.OFF_SIZE: W24SizeToleranceOffSize,
        W24SizeToleranceType.REFERENCE: W24SizeToleranceReference,
        W24SizeToleranceType.THEORETICALLY_EXACT:
        W24SizeToleranceTheoreticallyExact,
    }.get(type_)

    if class_ is None:
        return None

    return class_.parse_obj(raw)



""" HTTPS-part of the Werk24 client
"""
import base64
from types import TracebackType
from typing import Optional, Type
from urllib.parse import urlparse

import aiohttp
from pydantic import HttpUrl

from werk24.exceptions import (BadRequestException, RequestTooLargeException,
                               ResourceNotFoundException, ServerException,
                               UnauthorizedException,
                               UnsupportedMediaType)
from werk24.models.techread import W24PresignedPost

from .auth_client import AuthClient


class TechreadClientHttps:

    """ Translation map from the server response
    to the W24TechreadArchitectureStatus enum
    """

    def __init__(self, techread_server_https: str, techread_version: str):
        """ Intialize a new session with the https server

        Arguments:
            techread_server_https {str} -- Domain of the Techread https server
            techread_version {str} -- Techread Version
        """
        self._techread_server = techread_server_https
        self._techread_version = techread_version
        self._techread_session_https: Optional[aiohttp.ClientSession] = None
        self._auth_client: Optional[AuthClient] = None

    async def __aenter__(
            self
    ) -> 'TechreadClientHttps':
        """ Create a new HTTP session that is being used for the whole
        connection. Be sure to keep the session alive.

        Raises:
            RuntimeError  -- Raise when the developer enters the session
                without having called register_auth_client()

        Returns:
            TechreadClientHttps -- TechreadClientHttps version with active
                session
        """

        # make sure that we have an AuthClient
        if self._auth_client is None:
            raise RuntimeError(
                "You need to call register_auth_client() before you can start"
                + " the session")

        headers = {"Authorization": f"Bearer {self._auth_client.token}"}
        self._techread_session_https = aiohttp.ClientSession(headers=headers)
        return self

    async def __aexit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_value: Optional[BaseException],
        traceback: Optional[TracebackType]
    ) -> None:

        """ Close the session
        """
        if self._techread_session_https is not None:
            await self._techread_session_https.close()

    def register_auth_client(self, auth_client: AuthClient) -> None:
        """Register the reference to the authentication service

        Arguments:
            auth_client {AuthClient} -- Reference to Authentication
                client
        """
        self._auth_client = auth_client

    async def upload_associated_file(
        self,
        presigned_post: W24PresignedPost,
        content: Optional[bytes]
    ) -> None:
        """ Upload an associated file to the API.
        This can either be a technical drawing or a
        3D model. Potentially we will sometime extend
        this to also include cover pages.

        NOTE: the complete message size must not be
        larger than 10 MB

        Arguments:
            request_id {str} -- UUID4 request id that you obtained
                from the websocket

            filetype {str} -- filetype that we want to upload.
                currently supported: drawing, model

            content {bytes} -- content of the file as bytes

        Raises:

            BadRequestException: Raised when the request body
                cannot be interpreted. This normally indicates
                that the API version has been updated and that
                we missed a corner case. If you encounter this
                exception, it is very likely our mistake. Please
                get in touch!

            UnauthorizedException: Raised when the token
                or the requested file have expired

            ResourceNotFoundException: Raised when you are requesting
                an endpoint that does not exist. Again, you should
                not encounter this, but if you do, let us know.

            RequestTooLargeException: Raised when the status
                code was 413

            UnsupportedMediaTypException: Raised when the file you
                submitted cannot be read (because its media type
                is not supported by the API).

            ServerException: Raised for all other status codes
                that are not 2xx
        """
        # ignore if payload is empty
        if content is None:
            return

        # generate the form data by merging the presigned
        # fields with the file
        form = aiohttp.FormData()
        for key, value in presigned_post.fields_.items():
            form.add_field(key, value)
        form.add_field('file', content)

        # create a new fresh session that does not
        # carry the authentication token
        async with aiohttp.ClientSession() as sess:
            async with sess.post(presigned_post.url, data=form) as resp:

                # check the status code of the response and
                # raise the appropriate exception
                self._raise_for_status(presigned_post.url, resp.status)

    def _make_endpoint_url(
        self,
        subpath: str
    ) -> str:
        """ Make the endpoint url of the subpath.
        This will create a fully valid http url
        that can be used in the post and get requests

        Arguments:
            subpath {str} -- Path of the endpoint on
                the TechreadAPI

        Returns:
            str -- Fully qualified url including
                the server name and api version
        """
        return "https://{}/{}/{}".format(
            self._techread_server,
            self._techread_version,
            subpath)

    async def download_payload(self, payload_url: HttpUrl) -> bytes:
        """ Return the payload from the server

        Arguments:
            payload_url {HttpUrl} -- Url of the payload

        Raises:
            RuntimeError: Hard Error that is raised when
                the function is asked to download a payload
                from an untrusted source.
                This provides some sort of protection against
                payload-injection and token-theft. When you
                see this error showing up, you should
                definitely INVESTIGATE AND LET US KNOW
                IMMEDIATELY!!!
                Call all our numbers on a Sunday morning
                at 3am if it must be. Even if its Christmas
                and Easter on the same day.

            BadRequestException: Raised when the request body
                cannot be interpreted. This normally indicates
                that the API version has been updated and that
                we missed a corner case. If you encounter this
                exception, it is very likely our mistake. Please
                get in touch!

            UnauthorizedException: Raised when the token
                or the requested file have expired

            ResourceNotFoundException: Raised when you are requesting
                an endpoint that does not exist. Again, you should
                not encounter this, but if you do, let us know.

            RequestTooLargeException: Raised when the status
                code was 413

            UnsupportedMediaTypException: Raised when the file you
                submitted cannot be read (because its media type
                is not supported by the API).

            ServerException: Raised for all other status codes
                that are not 2xx
        Returns:
            bytes -- Payload
        """
        # send the get request to the endpoint
        try:
            response = await self._get(payload_url)

        # reraise the exceptions
        except (UnauthorizedException,  # pylint: disable=try-except-raise
                RequestTooLargeException,
                ServerException, BadRequestException,
                ResourceNotFoundException):
            raise

        # otherwise return the response text
        return base64.b64decode(await response.text())

    async def _get(
            self,
            url: str
    ) -> aiohttp.ClientResponse:
        """ Send a GET request request and return the
        response object. The method automatically
        injects the authentication token into the
        request.

        Arguments:
            url {str} -- URL that is to be requested

        Raises:
            BadRequestException: Raised when the request body
                cannot be interpreted. This normally indicates
                that the API version has been updated and that
                we missed a corner case. If you encounter this
                exception, it is very likely our mistake. Please
                get in touch!

            UnauthorizedException: Raised when the token
                or the requested file have expired

            ResourceNotFoundException: Raised when you are requesting
                an endpoint that does not exist. Again, you should
                not encounter this, but if you do, let us know.

            RequestTooLargeException: Raised when the status
                code was 413

            UnsupportedMediaTypException: Raised when the file you
                submitted cannot be read (because its media type
                is not supported by the API).

            ServerException: Raised for all other status codes
                that are not 2xx

        Returns:
            aiohttp.ClientResponse -- Client response for the get request
        """

        # ensure that the session was started
        if self._techread_session_https is None:
            raise RuntimeError(
                "You executed a command without opening a session")

        # send the request
        response = await self._techread_session_https.get(url)

        # check the status code of the response and
        # raise the appropriate exception
        try:
            self._raise_for_status(url, response.status)
        except (UnauthorizedException, ServerException) as exception:
            raise exception

        # if the call was successful, return
        return response

    @staticmethod
    def _raise_for_status(
            url: str,
            status_code: int
    ) -> None:
        """ Raise the correct exception depending on the
        status code

        Arguments:
            url {str} - - requested url
            status_code {int} - - response status code

        Raises:
            BadRequestException: Raised when the request body
                cannot be interpreted. This normally indicates
                that the API version has been updated and that
                we missed a corner case. If you encounter this
                exception, it is very likely our mistake. Please
                get in touch!

            UnauthorizedException: Raised when the token
                or the requested file have expired

            ResourceNotFoundException: Raised when you are requesting
                an endpoint that does not exist. Again, you should
                not encounter this, but if you do, let us know.

            RequestTooLargeException: Raised when the status
                code was 413

            UnsupportedMediaTypException: Raised when the file you
                submitted cannot be read(because its media type
                is not supported by the API).

            ServerException: Raised for all other status codes
                that are not 2xx

        """

        # raise a bad request exception if the status
        # code 400 was returned. This normally indicates
        # that the API has been updated and the integration
        # tests have missed a case
        if status_code == 400:
            raise BadRequestException()

        # raise an unauthorized exception if the
        # status code is
        # * 401 (Unauthorized) or
        # * 403 (Forbidden)
        if status_code in [401, 403]:
            raise UnauthorizedException()

        # NOTE: a 404 does not occur, as the
        # server does not want to tell you
        # whether the file does not exist
        # or whether your token is wrong.
        # Makes brute force attacks more expensive.
        # We deal with it anyway so we can change
        # in the future
        if status_code == 404:
            raise ResourceNotFoundException()

        # if the status code is 413, you have submitted
        # a file that is too large.
        if status_code == 413:
            raise RequestTooLargeException()

        # if the status code is 415, you have submitted
        # a file whose media type is not supported by the API
        if status_code == 415:
            raise UnsupportedMediaType()

        # If the resposne code is anything other
        # than unauthorized or 200 (OK), we trigger
        # a ServerException.
        if not 200 <= status_code <= 299:
            raise ServerException(
                f"Request failed '{url}' with code {status_code}")



from typing import Tuple
import abc
from decimal import Decimal
from enum import Enum
from typing import List, Optional

from pydantic import UUID4, BaseModel

from .feature import W24FeatureModel


class W24GDTCharacteristic(str, Enum):
    """ Enum of all possible Characteristics
    following ISO 1101.

    !!! note
        When the GDT frame only describes the location
        of a reference datum, the value DATUM_INDICATOR is used
    """
    FORM_STRAIGHTNESS = "⏤"
    FORM_FLATNESS = "⏥"
    FORM_CIRCULARITY = "○"
    FORM_CYLINDRICITY = "⌭"
    PROFILE_OF_SURFACE = "⌓"
    PROFILE_OF_LINE = "⌒"
    ORIENTATION_PERPENDICULARITY = "⟂"
    ORIENTATION_ANGULARITY = "∠"
    ORIENTATION_PARALLELISM = "∥"
    LOCATION_POSITION = "⌖"
    LOCATION_CONCENTRICITY = "◎"
    LOCATION_SYMMETRY = "⌯"
    RUNOUT_CIRCULAR = "↗"
    RUNOUT_TOTAL = "⌰"
    DATUM_INDICATOR = "[DATUM]"


class W24GDTFeatureAssociated(str, Enum):
    """ Enum of all associated tolerated features
    """
    MINIMAX = "Ⓒ"
    GAUSSIAN = "Ⓖ"
    MIN_CIRCUMSCRIBED = "Ⓝ"
    MAX_CIRCUMSCRIBED = "Ⓧ"
    TANGENT = "Ⓣ"


class W24GDTFeatureDerived(str, Enum):
    """ Enum of all derived features
    """
    PROJECTED = "Ⓟ"
    MEAN = "Ⓐ"


class W24GDTZoneCombination(str, Enum):
    """ Enum of all tolerance zone combinations
    """
    COMBINED = "CZ"
    SEPARATED = "SZ"


class W24GDTZoneShape(str, Enum):
    """ Enum of all zone shapes
    """
    DIAMETER = "Ø"
    DIAMETER_SPHERE = "S⌀"


class W24GDTZoneConstraint(str, Enum):
    """ Enum of the Zone Constraints
    """
    UNSPECIFIED_INCLINATION = "OZ"
    UNSPECIFIED_OFFSET = "VA"


class W24GDTDatum(BaseModel):
    """ Preliminary implementation of the GD&T Datum

    Attributes:
        blurb: Reference name. Typically: A,B,C ...
            Can also contain more complex names, e.g., (A-B-C-D)[CM]

    !!! note
        Future implementations might allow fine-grained
        access to the attributes of complex names:
        e.g., (A-B-C-D)[CM]
    """

    blurb: str


class W24GDTZoneOffset(BaseModel):
    """ Specified offset indicated by leading UZ...

    Attributes:
        blurb: blurb of the specified offset for human consumption.
            e.g., UZ+0.15, UZ-0.2, UZ+0.1:0.2
    """

    blurb: str


class W24GDTZoneValue(BaseModel):
    """ Preliminary defintion of the GDT Zone Value
    Future implementation will give access to the
    width and extend seperately

    Attributes:

        blurb: String representation for human consumption
            e.g., 0.05/12x10°

        width_min: Minimal width. Also used when no maximal width
            is defined.

        width_max: Optional maximal width.

        extend_quantity: Optional quantity of the spacing

        extend_shape: Optional shape of the extend

        extend_value: Optional extend value

        extend_angle: Optional angle of the extend
    """

    blurb: str

    width_min: Decimal

    width_max: Optional[Decimal]

    extend_quantity: Optional[int]

    extend_shape: Optional[W24GDTZoneShape]

    extend_value: Optional[Decimal]

    extend_angle: Optional[Decimal]


class W24GDTFilterType(str, Enum):
    """ Preliminary list of feature filters

    Filters remaining: RG, S, OH, SW, AB, CW
    """
    GAUSSIAN = "G"


class W24GDTFilter(BaseModel, abc.ABC):
    """ Abstract base class to describe feature filters

    Attributes:
        blurb: String representation of the file for human consumption

        filter_type: Filter Type to facilitate deserialization
    """

    blurb: str

    filter_type: W24GDTFilterType


# class W24GDTFilterG(W24GDTFilter):
#     """ Gaussian Feature filter without contraints
#     """
#     filter_type = W24GDTFilterType.GAUSSIAN

#     wavelength_min: float
#     """ Minimum wavelength in wavers/revolution
#     """

#     wavelength_max: float
#     """ Maximum wavelength in wavers/revolution
#     """


class W24GDTReferenceAssociation(str, Enum):
    """ Association of the Reference Element
    """

    MINIMAX = "C"
    """ Minimax (Tschebyschew-) Element without contraint
    """

    MINIMAX_EXTERNAL = "CE"
    """ Minimax (Tschebyschew-) Element with external
    constraint
    """

    MINIMAX_INTERNAL = "CI"
    """ Minimal (Tschebyschew-) Element with internal
    constraint
    """

    GAUSSIAN = "G"
    """ Gaussian Least Square Element without constraint
    """

    GAUSSIAN_EXTERNAL = "GE"
    """ Gaussian Least Square Element with external
    constraint
    """

    GAUSSIAN_INTERNAL = "GI"
    """ Gaussian Least Square Element with external
    constraint
    """

    MIN_CIRCUMSCRIBED = "N"
    """ Minimal circumscribed element
    """

    MAX_CIRCUMSCRIBED = "X"
    """ Maximal circumscribed element
    """


class W24GDTReferenceParameter(str, Enum):
    """ Parameter of the reference element
    """

    PEAK_VALUE = "P"
    """ Highest value
    """

    VALLEY_VALUE = "V"
    """ Lowest value
    """

    DEVISATION_SPAN = "T"
    """ Deviation span
    """

    STANDARD_DEVIATION = "Q"
    """ Root mean square deviation
    """


class W24GDTMaterialCondition(str, Enum):
    """ Enum for Material Conditions
    """

    MAXIMUM = "Ⓜ"
    """ Maximum material condition (MMC)
    """

    MINIMUM = "Ⓛ"
    """ Mimimal material condition (LMC)
    """

    RECIPROCITY = "Ⓡ"
    """ Reciprocity
    """


class W24GDTState(str, Enum):
    """ Enum for the State
    """
    FREE = "Ⓕ"


class W24GDTFrame(BaseModel):
    """ Representation of the Geometric Dimensioning
    and Toleration frame

    Attributes:
        gdt_id: Unique id of the GDT

        blurb: String representation of the label for human consumption
            e.g., [⌖|⌀0.3Ⓜ|A|B|C]

        characteristic: Section for geometric characteristic e.g.: ⌓

        zone_shape: Tolerance zone shape, e.g, S⌀

        zone_value: GDT value: e.g., 0.03
            Is optional to support Datum Feature Indicators

        zone_combinations: Ordered list of zone combinations, e.g., CZ, SZ

        zone_offset: Optional specified offset, e.g., UZ-0.2

        zone_constraint: Optional zone constraint: e.g., OZ, VA

        feature_filter: Optional feature filter

        feature_associated: Associated toleraced feature

        feature_derived: Derived Feature

        reference_association: Reference element association

        reference_parameter: Reference element parameter

        material_condition: Material condition

        state: FREE or None

        data: Ordered list of data

    """
    gdt_id: Optional[UUID4] = None

    blurb: str

    characteristic: W24GDTCharacteristic

    zone_shape: Optional[W24GDTZoneShape] = None

    zone_value: Optional[W24GDTZoneValue]

    zone_combinations: List[W24GDTZoneCombination] = []

    zone_offset: Optional[W24GDTZoneOffset] = None

    zone_constraint: Optional[W24GDTZoneConstraint] = None

    feature_filter: Optional[W24GDTFilter] = None

    feature_associated: Optional[W24GDTFeatureAssociated] = None

    feature_derived: Optional[W24GDTFeatureDerived] = None

    reference_association: Optional[W24GDTReferenceAssociation] = None

    reference_parameter: Optional[W24GDTReferenceParameter] = None

    material_condition: Optional[W24GDTMaterialCondition] = None

    state: Optional[W24GDTState] = None

    data: List[W24GDTDatum] = []


class W24GDT(W24FeatureModel):
    """ Parent object for Geometric Dimensionsing and Toleration
    Frames, attaching them to the physical location on the drawing.

    Attributes:
        frame: Representation of the GDT frame

    """
    # !!! DEPRECATED
    bounding_polygon: List[Tuple[float, float]]

    frame: W24GDTFrame

    # measure_label: Optional[W24MeasureLabel] = None
    """ Optional size dimension, typically annotated
    above the feature control frame.

    Example:
        ⌀12.0 + /- 0.1
        [⊥|⌀0.1Ⓜ| A]
    """

    # sectional_plane
    """ Definition of Sectional Plane Indicator will
    follow.
    Example: ⟨[∥| A]
    """  # pylint: disable=pointless-string-statement

    # orientation_plane
    """ Definition of the Orientational Plane Indicator
    will follow
    Example: ⟨[∥| A]⟩
    """  # pylint: disable=pointless-string-statement

    # direction_feature
    """ Defintion of the Direction Feature Indicator
    will follow
    Example: ←[∥| A]
    """  # pylint: disable=pointless-string-statement

    # collection_plane
    """ Defintion of the Collection Plane Indicator
    will follow
    Exmple: ◯[∥| B]
    """  # pylint: disable=pointless-string-statement



from enum import Enum


class W24FileFormatVariantCAD(str, Enum):
    """ List of the supported CAD file formats

    !!! note
        At this stage, the API only supports DXF
        as output format. Additional formats are available
        upon request.
    """
    DXF = "DXF"


class W24FileFormatThumbnail(str, Enum):
    """ List of supported File Formats in which
    the Thumbnail can be supplied.

    !!! note
        At this stage, the API only supports JPEG.
        This prepares the support for additional format
        (e.g. TIFF)
    """
    JPEG = "JPEG"



import sys
from werk24.techread_client import LicenseError, W24TechreadClient

def make_client() -> W24TechreadClient:
    """ Make the client. 
    
    This will automatically
    fetch the authentication information
    from the environment variables. We will
    provide you with separate .env files for
    the development and production environments

    Returns:
        W24TechreadClient: Client instance
    """
    
    try:
        return W24TechreadClient.make_from_env()

    # If a license error occured, let the user know.
    # NOTE: This will not catch deactivated users.
    # They will only learn about their status once
    # they send a request.
    except LicenseError as exception:
        print(f"LICENSE ERROR: {exception}")
        sys.exit()


""" W24Client Module

DESCRIPTION
    The module contains everything that is needed to
    communicate with the W24 API - allowing you
    to interpret the contents of your technical drawings.

AUTHOR
    Jochen Mattes (Werk24)

EXAMPLE
    # obtain the thumbnail of a page
    drawing_bytes = open(...,"r").read()
    client = W24TechreadClient.make_from_env()
    await client.read_drawing_with_hooks(
        drawing_bytes,
        [Hook(
                ask=W24AskPageThumbnail(),
                callback=lambda msg: print("Received Thumbnail of Page")
        ]))
"""
import logging
import os
import uuid
from asyncio import gather, iscoroutinefunction
from types import TracebackType
from typing import (
    AsyncGenerator,
    AsyncIterator,
    Callable,
    Dict,
    List,
    Optional,
    Type,
    Union,
)

import dotenv
from pydantic import UUID4, BaseModel

from werk24.auth_client import AuthClient
from werk24.exceptions import (
    BadRequestException,
    LicenseError,
    RequestTooLargeException,
    ServerException,
    UnsupportedMediaType,
)
from werk24.models.ask import W24Ask
from werk24.models.techread import (
    W24TechreadAction,
    W24TechreadBaseResponse,
    W24TechreadException,
    W24TechreadExceptionLevel,
    W24TechreadExceptionType,
    W24TechreadInitResponse,
    W24TechreadMessage,
    W24TechreadMessageSubtype,
    W24TechreadMessageType,
    W24TechreadRequest,
)
from werk24.techread_client_https import TechreadClientHttps
from werk24.techread_client_wss import TechreadClientWss

# make the logger
logger = logging.getLogger('w24_techread_client')

ENVIRONS = [
    "W24TECHREAD_SERVER_HTTPS",
    "W24TECHREAD_SERVER_WSS",
    "W24TECHREAD_VERSION",
    "W24TECHREAD_AUTH_CLIENT_ID",
    "W24TECHREAD_AUTH_CLIENT_SECRET",
    "W24TECHREAD_AUTH_IDENTITY_POOL_ID",
    "W24TECHREAD_AUTH_USER_POOL_ID",
    "W24TECHREAD_AUTH_USERNAME",
    "W24TECHREAD_AUTH_PASSWORD",
    "W24TECHREAD_AUTH_REGION"
]
""" List of the environment variables used by the
client """


EXCEPTION_MAP = {
    RequestTooLargeException:
        W24TechreadExceptionType.DRAWING_FILE_SIZE_TOO_LARGE,
    BadRequestException:
        W24TechreadExceptionType.DRAWING_FILE_SIZE_TOO_LARGE
}
""" Map to translate the local exceptions to official
W24Exceptions. This allows us to mock consistent responses
even when the files are rejected before they reach the API
"""

DEFAULT_AUTH_REGION = "eu-central-1"
DEFAULT_SERVER_HTTPS = "techread.w24.io"
DEFAULT_SERVER_WSS = "techread-ws.w24.io"
DEFAULT_VERSION = "v1"


LICENSE_ERROR_TEXT = """

////////////////////////////////////////////////////////////////////////////////

,.'xWWx'.,kWWd'.;0O,..'',,,,:Ox'..',,'',:lOWNo..:0MXd,.'lKWMNOc'.,;;'....';,....
l..lXK:...cXKc..oNO,..o0KKKKXNx'.'dKKOc...;0No..;OOc..:kNMMXl'.'oOOOO:..;kKd'...
k,.,kx'...,kx,.;OMO,..:dddxkXWx'.'xXX0c...:0No..,:,.'oXWMMMO,..,lc;kXl.:kKNk'...
Xc..:;.;l,.::..lNMO,..,ccccoKWx'..,;,'.';l0WNo.......lKMMMMO,....;x0o,:00ONO;...
Wk'....dKl....,OMMO,.'dNWWWWWWx'.'okl'.,:xNMNo..'ld:..:OWMMO,..'oKXkl;cxxOWKc...
MKc...;0WO,...lXMMO,..;oooooxKx'.,OMNd'..,dNNo..;0MKl'.,dNMO,..,oxxddc'..;oc..;x
MWd'..lNMXl..'xWMMO,........;Ox'.,OMMNo...,kNo..:0MMKc..,xWO,................c0W

////////////////////////////////////////////////////////////////////////////////

Thank you for your interest ist Werk24.

_General Information_
Werk24 is a commercial company focusing on information extraction from
Technical Documents. The client you are running right now allows you to to
communicate with our server infrastructure.

_License File_
We were unable to locate a license file, please schedule a first meeting with
us to learn about the possibilities of our Technology:


        >>> https://werk24.io/schedule-a-call?w24cli_license_error <<<


Thank you.

////////////////////////////////////////////////////////////////////////////////
"""


class Hook(BaseModel):
    """ Small Object to keep the callback requests.
    You can either register a callback request to an
    ask or a message_type.

    If you register an ask, be sure to use a complete W24Ask
    definition; not just the ask type.
    """
    message_type: Optional[W24TechreadMessageType]
    message_subtype: Optional[W24TechreadMessageSubtype]
    ask: Optional[W24Ask]
    function: Callable


class W24TechreadClient:
    """ Simple W24Client that allows you to use
    learn more about the content on your Technical
    Drawings.
    """

    def __init__(
        self,
        techread_server_https: str,
        techread_server_wss: str,
        techread_version: str,
        development_key: str = None
    ):
        """ Initialize a new W24TechreadClient. If you wonder
        about any of the attributes, have a look at the .env
        file that we provided to you. They contain all the
        information that you will need.

        Arguments:
            techread_server_https {str} -- domain name that
                is being used by the https client

            techread_server_wss {str} -- domain name that
                is being used by the websocket client

            techread_version {str} -- version that you want to
                connect to

            development_key {str} -- key that allows you to submit
                your request to one of the internal architectures.
                You can try guessing or bruteforcing this key;
                we'll just charge you for every request you submit and
                transfer the money to the holiday bonus account.
        """

        # save the development_key
        self._development_key = development_key

        # Create an empty reference to the authentication
        # service (currently AWS Cognito)
        self._auth_client: Optional[AuthClient] = None

        # Initialize an instance of the HTTPS client
        self._techread_client_https = TechreadClientHttps(
            techread_server_https, techread_version)

        # Initialize an instance of the WEBSOCKET client
        self._techread_client_wss = TechreadClientWss(
            techread_server_wss, techread_version)

    async def __aenter__(
            self
    ) -> 'W24TechreadClient':
        """ Create the HTTPS and WSS sessions

        Raises:
            RuntimeError: Exception is raised if
                you tried to enter a session before
                calling the register() method

        Returns:
            W24TechreadClient -- Version of self with
                active sessions
        """

        # enter the https session
        await self._techread_client_https.__aenter__()

        # enter the wss session
        await self._techread_client_wss.__aenter__()

        # return the "entered" version of self
        return self

    async def __aexit__(
            self,
            exc_type: Optional[Type[BaseException]],
            exc_value: Optional[BaseException],
            traceback: Optional[TracebackType]
    ) -> None:

        """ Ensure that the sessions are closed
        """

        # close the HTTPS session
        await self._techread_client_https.__aexit__(
            exc_type, exc_value, traceback)

        # close the WSS session
        await self._techread_client_wss.__aexit__(
            exc_type, exc_value, traceback)

    def register(
        self,
        cognito_region: str,
        cognito_identity_pool_id: str,
        cognito_user_pool_id: str,
        cognito_client_id: str,
        cognito_client_secret: str,
        username: str,
        password: str
    ) -> None:
        """
        Register with the authentication
        service (i.e., lazy login)

        Arguments:
            cognito_region {str} -- Physical region
            cognito_identity_pool_id {str} -- identity pool of W24
            cognito_client_id {str} -- the client id of your application
            cognito_client_secret {str} -- the client secret of your
                application
            username {str} -- the username with which you want to register
            password {str} -- the password with which you want to register
        """
        # create an client instance to connect
        # to the authentication service
        self._auth_client = AuthClient(
            cognito_region,
            cognito_identity_pool_id,
            cognito_user_pool_id,
            cognito_client_id,
            cognito_client_secret)

        # register username and password
        self._auth_client.register(username, password)

        # tell the techread clients about it
        self._techread_client_https.register_auth_client(self._auth_client)
        self._techread_client_wss.register_auth_client(self._auth_client)

        # ensure that we have a token
        try:
            self._auth_client.login()  # type: ignore
        except AttributeError as exc:
            raise RuntimeError(
                "No connection to the authentication service was " +
                "established. Please call register()") from exc

    @property
    def username(self) -> Optional[str]:
        """ Make the username accessible to the CLI and GUI

        Returns:
            str: username of the currently registered user
        """
        if self._auth_client is None:
            return None

        return self._auth_client.username

    async def read_drawing(
        self,
        drawing: bytes,
        asks: List[W24Ask],
        model: bytes = None,
        max_pages: int = 1,
        drawing_filename: Optional[str] = None,
        sub_account: Optional[UUID4] = None
    ) -> AsyncIterator[W24TechreadBaseResponse]:
        """ Send a Technical Drawing to the W24 API to have it automatically
        interpreted and read. The API will return


        Arguments:
            drawing (bytes): binary representation of a technical drawing.
                Please refer to the API - documentation to learn which mime
                types are supported.

            model (bytes): binary representation of the 3d model (typically
                step)
                Please refer to the API - documentation to learn which mime
                types are supported. Defaults to None.

            asks (List[W24Ask]): List of Asks that are requested from the API.
                They must derive from the W24Ask object. Refer to the API
                documentation for a full list of supported W24AskTypes

            max_pages (int): Maximum number of pages that are being processed
                of the submitted file. This protects platform users from
                costly requests caused by a user uploading a single file with
                many pages.

            drawing_filename (str|None): Optional information about the
                filename of the drawing. Frequently this contains information
                about the drawing id and you can make that information
                available to us through this parameter. If you don't know the
                filename, don't worry, it will still work.

            sub_account (UUID4|None): Optional specification of the sub-account
                that the request should be attributed to. Sub-accounts allow
                you to manage several customers at the same time and receiving
                separate positions on the monthly invoice.

        Yields:
            W24TechreadMessage -- Response object obtained from the API
                that indicates the state of your request. Be sure to pass this
                to the read_drawing_listen method

        Raises:
            DrawingTooLarge -- Exception is raised when the drawing was too
                large to be processed. At the time of writing. The upload
                limit lies at 6 MB (including overhead).

            UnsupportedMediaType -- Exception is raised when the drawing or
                model is submitted in any data type other than bytes.
        """

        # quickly check whether the input type is bytes. If it is string,
        # the presigned-AWS post interestingly returns a 403 error_code
        # without additional information. We want to inform the caller
        # that they submitted the wrong data type.
        # See Github Issue #13
        if not isinstance(drawing, bytes):
            raise UnsupportedMediaType("Drawing bytes requires 'bytes' type")

        # the same is true for the model bytes
        if model is not None and not isinstance(model, bytes):
            raise UnsupportedMediaType("Model bytes requires 'bytes' type")

        # give us some debug information
        logger.info("API method read_drawing() called")

        # tell us when a development key is being used
        if self._development_key:  # pragma: no cover
            logger.info("Using development key %s***",
                        self._development_key[:8])

        # send the initiation command
        init_response = await self._send_command_init(
            asks,
            max_pages,
            drawing_filename,
            sub_account)

        # stop if the response is unsuccessful.
        if not init_response.is_successful:
            yield init_response
            return

        # upload drawing and model. We can do that in parallel.
        # If your user uploads them separately, you could also
        # upload them separately to Werk24.
        try:
            await gather(
                self._techread_client_https.upload_associated_file(
                    init_response.drawing_presigned_post,
                    drawing),
                self._techread_client_https.upload_associated_file(
                    init_response.model_presigned_post,
                    model))

        # explicitly reraise the exception if the payload is too
        # large
        except (BadRequestException, RequestTooLargeException) as exception:
            async for message in self._trigger_asks_exception(asks, exception):
                yield message
            return

        # Tell Werk24 that all the files have been uploaded
        # correctly and the reading process can be started.
        #
        # NOTE: you will only be able to start the reading
        # process from the websocket connection that
        # initiated the request. If you want to run a
        # stateless-system that separates the initialization
        # from the upload and read stages, you'll need to
        # find a way of handing over the tcp connection :)
        # PS: The AWS API Gateway for websockets might help you
        # here.
        async for message in self._send_command_read():
            yield message

    async def _send_command_init(
        self,
        asks: List[W24Ask],
        max_pages: int,
        drawing_filename: Optional[str],
        sub_account: Optional[UUID4]
    ) -> W24TechreadInitResponse:
        """ Send the initiation command to the backend
        and return the associated response

        This achieves two things:
        1. The server has a couple of 100ms to
           reserves some resources for you, and
        2. The server will create a new request_id
           that you will need when uploading the
           associated files

        Args:
            asks (List[W24Ask]): List of asks that
                we are asking for in this request

            max_pages (int): Maximum number of pages
                that shall be read.

            drawing_filename (Optional[str]): Optional
                filename of the drawing.
        """
        request = W24TechreadRequest(
            asks=asks,
            development_key=self._development_key,
            max_pages=max_pages,
            drawing_filename=drawing_filename,
            sub_account=sub_account)

        await self._techread_client_wss.send_command(
            W24TechreadAction.INITIALIZE.value,
            request.json())

        response = await self._techread_client_wss.recv_message()
        logger.info("Received request_id %s", response.request_id)
        return W24TechreadInitResponse.parse_obj(response.payload_dict)

    async def _send_command_read(
        self
    ) -> AsyncGenerator[W24TechreadMessage, None]:
        """ Send the request request to the backend
        and yield the resulting messages

        Yields:
            W24TechreadMessage: Receiving messages
        """

        # submit the request the to the API
        await self._techread_client_wss.send_command(
            W24TechreadAction.READ.value,
            "{}")
        logger.info("Techread request submitted")

        # Wait for incoming messages from the server.
        # They will tell you when the individual
        # asks become available. The socket returns
        # strings of jsonified W24TechreadMessage objects.
        #
        # The loop will stop when the websocket is
        # closed
        async for message in self._techread_client_wss.listen():

            # check whether we need to download something
            if message.payload_url is not None:
                message.payload_bytes = await self._techread_client_https \
                    .download_payload(message.payload_url)

            # return the message to the caller for immediate
            # consumption
            yield message

    @staticmethod
    async def _trigger_asks_exception(
        asks: List[W24Ask],
        exception_raw: Union[BadRequestException, RequestTooLargeException]
    ) -> AsyncGenerator[W24TechreadMessage, None]:
        """ Trigger exceptions for all the submitted asks.
        This helps us to mock consistent exception handling
        behavior even when the files are rejected before they
        reach the API.

        Args:
            asks (List[W24Ask]): List of all submitted asks
            exception (RequestTooLargeException): Exception
                that shall be pushed

        Yields:
            W24TechreadMessage: Exception message
        """

        # get the exception type from the MAP
        try:
            exception_type = EXCEPTION_MAP[type(exception_raw)]

        # if we see an exception that we were not supposed
        # to handle, there must have been a developer passing
        # a new exception type. Let's tell her by rasing
        # a runtime error
        except KeyError:
            raise RuntimeError(
                f"Unknown exception type passed: {type(exception_raw)}")

        # translate the exception into an official exception
        exception = W24TechreadException(
            exception_level=W24TechreadExceptionLevel.ERROR,
            exception_type=exception_type)

        # then yield one message for each of the requested asks
        for cur_ask in asks:
            yield W24TechreadMessage(
                request_id=uuid.uuid4(),
                message_type=W24TechreadMessageType.ASK,
                message_subtype=cur_ask.ask_type,
                exceptions=[exception])

    @ staticmethod
    def _get_license_environs(
        license_path: Optional[str]
    ) -> Dict[str, str]:
        """ Get the environment variables
        Where we either select the variables from the license
        files. If that fails we fall back to the true environment
        variables.

        NOTE: We do not want to mix the sources.

        Args:
            license_path (Optional[str]): Path of the license files

        Returns:
            Dict[str,str]: Key, Value pairs for the environment variables
        """

        # Mimick the old default value of .werk24
        if license_path is None and os.path.exists(".werk24"):
            license_path = ".werk24"  # pragma: no cover

        # First priority: look for the local license path
        if license_path is not None:
            if os.path.exists(license_path):
                environs_raw = {
                    k: v
                    for k, v in dotenv.dotenv_values(license_path).items()
                    if v is not None}

            # if the caller defined a license path, but it does not
            # exist, raise the exception
            else:
                raise LicenseError(LICENSE_ERROR_TEXT)

        # Second priority: use the environment variables
        else:
            environs_raw = dict(os.environ)

        # filter the environment variables to only include the
        # ones that are relevant to us and return
        try:
            return {cur_key: environs_raw[cur_key] for cur_key in ENVIRONS}
        except KeyError:
            raise LicenseError(LICENSE_ERROR_TEXT)

    @classmethod
    def make_from_env(
        cls,
        license_path: Optional[str] = None,
        auth_region: Optional[str] = None,
        server_https: Optional[str] = None,
        server_wss: Optional[str] = None,
        version: Optional[str] = None
    ) -> "W24TechreadClient":
        """ Small helper function that creates a new
        W24TechreadClient from the environment info.

        Arguments:
            license_path:{Optional[str]} -- path to the License file.
                By default we are looking for a .werk24 file in the current
                cwd. If argument is set to None, we are not loading any
                file and relying on the ENVIRONMENT variables only

            auth_region: {Optional[str]} -- AWS Region of the Authentication
                Service.
                Takes priority over environ W24TECHREAD_AUTH_REGION and
                DEFAULT_AUTH_REGION

            server_https: {Optional[str]} -- HTTPS endpoint of the Werk24 API.
                Takes priority over environ W24TECHREAD_SERVER_HTTPS and
                DEFAULT_SEVER_HTTPS

            version: {Optional[str]} -- Version of the Werk24 API.
                Takes priority over environ W24TECHREAD_VERSION and
                DEfAULT_VERSION

        Raises:
            FileNotFoundError -- Raised when you pass a path to a license file
                that does not exist
            UnauthorizedException -- Raised when the credentials were not
                accepted by the API

        Returns:
            W24TechreadClient -- The techread Client
        """

        # get the license variables from the environment variables and
        # the license file.
        environs = cls._get_license_environs(license_path)

        # define a small helper function that finds the first valid
        # value in the supplied list of possible values
        def pick_env(
            var: Optional[str],
            env_key: str,
            default: str
        ) -> str:
            return var or environs.get(env_key) or default

        # then make sure we use the correct priorities
        auth_region = pick_env(
            auth_region,
            'W24TECHREAD_AUTH_REGION',
            DEFAULT_AUTH_REGION)

        server_https = pick_env(
            server_https,
            'W24TECHREAD_SERVER_HTTPS',
            DEFAULT_SERVER_HTTPS)

        server_wss = pick_env(
            server_wss,
            'W24TECHREAD_SERVER_WSS',
            DEFAULT_SERVER_WSS)

        version = pick_env(
            version,
            'W24TECHREAD_VERSION',
            DEFAULT_VERSION)

        # get the variables from the environment and ensure that they
        # are set. If not, raise an exception
        try:

            # create a reference to the client
            client = W24TechreadClient(server_https, server_wss, version)

            # register the credentials. This will in effect
            # only set the variabels in the authorizer. It will
            # not trigger a network request
            client.register(
                auth_region,
                environs['W24TECHREAD_AUTH_IDENTITY_POOL_ID'],
                environs['W24TECHREAD_AUTH_USER_POOL_ID'],
                environs['W24TECHREAD_AUTH_CLIENT_ID'],
                environs['W24TECHREAD_AUTH_CLIENT_SECRET'],
                environs['W24TECHREAD_AUTH_USERNAME'],
                environs['W24TECHREAD_AUTH_PASSWORD'])

        except KeyError:
            raise LicenseError(LICENSE_ERROR_TEXT)

        # return the client
        return client

    async def read_drawing_with_hooks(
        self,
        drawing_bytes: bytes,
        hooks: List[Hook],
        max_pages: int = 1,
        drawing_filename: Optional[str] = None,
        sub_account: Optional[UUID4] = None
    ) -> None:
        """ Send the drawing to the API (can be PDF or image)
        and register a number of callbacks that are triggered
        once the asks become available.

        Arguments:
            drawing_bytes {bytes} -- Technical Drawing as Image or PDF
            hooks {List[Hook]} -- List of Callback you want to obtain

        Raises:
            ServerException -- Raised when the server returns an ERROR
                message
        """

        # filter the callback requests to only contain
        # the ask types
        asks_list = [
            cur_ask.ask
            for cur_ask in hooks
            if cur_ask.ask is not None
        ]

        try:
            # send out the request and make a generator
            # that triggers when the result of an ask
            # becomes available
            async for message in self.read_drawing(
                    drawing_bytes,
                    asks_list,
                    max_pages=max_pages,
                    drawing_filename=drawing_filename,
                    sub_account=sub_account):

                await self._call_hooks_for_message(message, hooks)

        # explicitly reraise server exceptions
        except ServerException:  # pylint: disable=try-except-raise
            raise

    async def _call_hooks_for_message(
            self,
            message: W24TechreadMessage,
            hooks: List[Hook]
    ) -> None:
        """ Find the correct hook for the read response and
        call the corresponding hook.

        Arguments:
            message {W24TechreadMessage} -- Message returned from the
                read_drawing method

            hooks {List[Hook]} -- List of hooks from which we need to
                pick the suited one

        Raises:
            ServerException: raised when the server returns an ERROR
                message
        """

        # get the hook function that corresponds to the message
        hook_function = self._get_hook_function_for_message(message, hooks)

        # if no hook is registered, ignore
        if hook_function is None:
            return

        # if the hook_function is not callable, we want to warn the user,
        # rather than throwing an exception
        if not callable(hook_function):
            logger.warning(
                "You registered a non-callable trigger of type '%s' with "
                "the message_type '%s'. Please make sure that you are using "
                "a Callable (e.g, def or lambda)",
                type(hook_function),
                message.message_type)
            return

        # if everything went well, we call the trigger with
        # the message as payload. Be sure to call the
        # function asymmetrically if supported
        if iscoroutinefunction(hook_function):
            await hook_function(message)
        else:
            hook_function(message)

    @staticmethod
    def _get_hook_function_for_message(
            message: W24TechreadMessage,
            hooks: List[Hook]
    ) -> Optional[Callable]:
        """ Get the hook function that corresponds to the message
        type.

        Arguments:
            message {W24TechreadMessage} -- Message returned from the
                read_drawing method

            hooks {List[Hook]} -- List of hooks from which we need to
                pick the suited one

        Returns:
            Optional[Callable] -- Hook function that should be called
        """

        # because we allow the user to define the ask in the hook,
        # we need to make some extra effort when filtering
        if message.message_type == W24TechreadMessageType.ASK:
            def hook_filter(hook: Hook) -> bool:
                return hook.ask is not None \
                    and message.message_subtype.value \
                    == hook.ask.ask_type.value

        # if the message is of any other type, we just need to
        # compare the the message_type and message_subtype
        else:
            def hook_filter(hook: Hook) -> bool:
                return hook.message_type is not None \
                    and hook.message_subtype is not None \
                    and message.message_type == hook.message_type \
                    and message.message_subtype == hook.message_subtype

        # return the first positive case
        for cur_hook in filter(hook_filter, hooks):
            return cur_hook.function

        # if we are still here, we have an unknown message type, which
        # probably is being caused by an API update. We want to ensure
        # that the user is being informed, but we do not want to break
        # the existing functionality -> warning
        logger.warning(
            "Ignoring message of type %s:%s - no hook registered",  # noqa
            message.message_type,
            message.message_subtype)
        return None



""" Definition of the Thread support structures

Author: Jochen Mattes - Werk24
"""
import abc
from decimal import Decimal
from enum import Enum
from typing import List, Dict, Any, Optional, Type

from pydantic import BaseModel, validator

from werk24.models.feature import W24FeatureModel

from .unit import W24UnitLength
from .size import W24SizeTolerance
from .gender import W24Gender

class W24ThreadType(str, Enum):
    """ Enum for the individual thread types

    NOTE: UTS_COARSE, UTS_FINE, UTS_EXTRAFINE and UTS_SPECIAL
        as individual types will be deprecated. Their information
        individual `threads_per_inch` information will be stored
        in the corresponding variable.
    """
    ISO_METRIC = "ISO_METRIC"
    WHITWORTH = "WHITWORTH"
    UTS = "UTS"

    # !!! DEPRECATED
    UTS_COARSE = "UTS_COARSE"

    # !!! DEPRECATED
    UTS_FINE = "UTS_FINE"

    # !!! DEPRECATED
    UTS_EXTRAFINE = "UTS_EXTRAFINE"

    # !!! DEPRECATED
    UTS_SPECIAL = "UTS_SPECIAL"


class W24ThreadHandedness(str, Enum):
    """ Enum describing the direction of the
    thread
    """
    LEFT = "LEFT"
    RIGHT = "RIGHT"

class W24ThreadMultiStart(BaseModel):
    """ Multi-Start Thread Information

    A common example of multi-start threads are the cap on a plastic
    water bottle. The cap will screw on quickly because a multi-start thread.

    Attributes:

        thread_lead (Decimal, optional): Thread lead for multiple
            starts

        number_of_starts (int): Number of intertwined threads running
            in parallel to one another.

    """
    thread_lead: Optional[Decimal]
    number_of_starts: int = 1


class W24Thread(BaseModel, abc.ABC):
    """ Abstract Base Class for all Threads

    Attributes:
        blurb (str): String representation of the thread for human consumption

        diameter (Decimal): Diameter of the thread

        diameter_preference: Not all Threads are made equal. The standardization
            bodies assign different preferences to the different thread diameters.

        unit (W24UnitLength): Length unit of the diameter. This allows you
            to differentiate between inch and millimeter (UTS or ISO METRIC).
            NOTE: Whitworth inch are not 2.54mm long! (for historic reasons).
            Use the whitworth_size if you need to have the Whitworth inches.

        thread_type: thread type to facilitate deserialization

        pitch: Pitch of the thread in mm. Normed range: 0.25 - 9mm
            The value will always be set - even if the pitch is give implicitly
            (e.g. M10c) or when the pitch is given in threads per inch.
            When the Pitch is ambiguous, this value be the median pitch

        threads_per_inch: Pitch and Threads per inch are interchangeable
            concepts. Depending on your geographic focus, the one or the
            other concept is more suitable. We convert ISO pitches to
            UTS threads per inch and vice versa, so that you do not have
            to handle the conversion.

        handedness: Left of right-handedness of the thread.
            This will be RIGHT unless explicitly described as LEFT
            in the drawing.

        length (Decimal, optional): Length of the Thread if specified
            on the label (e.g., M15x1x20). None otherwise.
            NOTE: If the thread length is not specified in the label, you can
            still obtain the information from the W24ThreadElement
            NOTE: If the thread label specifies both a thread length and a
            bore length, only the thread length will be included here. The
            bore length can be found in the W24MeasureElement.depth attribute

    """
    blurb: str

    diameter: Decimal

    unit: W24UnitLength = W24UnitLength.MILLIMETER

    thread_type: W24ThreadType

    pitch: Decimal

    threads_per_inch: Decimal

    handedness: W24ThreadHandedness = W24ThreadHandedness.RIGHT

    multi_start: W24ThreadMultiStart

    length: Optional[Decimal]


class W24ThreadISOMetricThreadEngagementClass(str, Enum):
    """Thread Engagement in accordance with ISO 965-1:1998
    """
    SHORT = "SHORT"
    NORMAL = "NORMAL"
    LONG = "LONG"


class W24ThreadISOMetric(W24Thread):
    """ Metric ISO Thread following ISO 1502

    Supports
    * DIN 14-1 threads (i.e., diameter < 1mm)
    * DIN 13-1 threads (i.e., diameter 1..68mm)
    * DIN 13-2 to DIN 13-10 threads (i.e., diameter 1..1000mm)
    * DIN 158-1 (i.e., cone-shaped threads)

    NOTE: ISO 965-1:1998 allows for the specification of fits
    between threaded parts. This is considered by giving access
    to all internal and external tolerance parameters.

    Attributes:
        inner_major_diameter_tolerance (str): Tolerance Class
            for the major diameter in accordance with ISO 965-1.
            None if the major diameter tolerance class is neither
            defined in the standard not explicitly specified

        inner_pitch_diameter_tolerance (str): Tolerance Class
            for the pitch diameter in accordance with ISO 965-1.
            None if the pitch diameter tolerance class is neither
            defined in the standard not explicitly specified

        outer_major_diameter_tolerance (str): Corresponding
            tolerance class for the major diameter of the
            outer thread

        outer_pitch_diameter_tolerance (str): Corresponding
            tolerance class for the pitch diameter of the outer thread.

    """
    thread_type = W24ThreadType.ISO_METRIC

    female_major_diameter_tolerance: Optional[W24SizeTolerance]
    female_pitch_diameter_tolerance: Optional[W24SizeTolerance]
    male_major_diameter_tolerance: Optional[W24SizeTolerance]
    male_pitch_diameter_tolerance: Optional[W24SizeTolerance]


class W24ThreadUTS(W24Thread):

    """ Unified Thread Standard (UTS) base class for
    * UNC - Unified National Coarse Thread
    * UNF - Unified National Fine Thread
    * UNEF - Unified National Extrafine Thread

    Attributes:
        uts_size: UTS size as string representation.
            Threads with a diameter < 0.25 inch are written with a leading '#'.
            Threads with a diameter >= 0.25 inch are represented as fractions
                with a tailing '"'
            Examples: #0, 1 3/4"

        uts_series (str): UTS Forma n series following ASME B1.1 2003.
            Valid values include UNC UNF, UNRF, UNF, UN

        threads_per_inch: Threads per inch. The Decimal (rather than int)
            is chosen to support non-conventional threads.

        tolerance_class: Tolerance class. Options:
            * 1A, 2A, 3A for external threads
            * 1B, 2B, 3B for internal threads
    """
    thread_type = W24ThreadType.UTS

    uts_size: str
    uts_series: str

    tolerance_class: str


class W24ThreadUTSCoarse(W24ThreadUTS):
    """ Unified National Coarse Thread

    NOTE: will be deprecated in favor of W24ThreadUTS
    """
    thread_type = W24ThreadType.UTS_COARSE


class W24ThreadUTSFine(W24ThreadUTS):
    """ Unified National Fine Thread

    NOTE: will be deprecated in favor of W24ThreadUTS
    """
    thread_type = W24ThreadType.UTS_FINE


class W24ThreadUTSExtrafine(W24ThreadUTS):
    """ Unified National Extrafine Thread

    NOTE: will be deprecated in favor of W24ThreadUTS
    """
    thread_type = W24ThreadType.UTS_EXTRAFINE


class W24ThreadUTSSpecial(W24ThreadUTS):
    """ Unified National Special Thread

    NOTE: will be deprecated in favor of W24ThreadUTS
    """
    thread_type = W24ThreadType.UTS_SPECIAL


class W24ThreadWhitworth(W24Thread):
    """ Whitworth Thread following ISO 228-1

    Attributes:
        whitworth_size: Size number (for historic reasons
            not proportional to size)

        tolerance_class: Tolerance field

    """
    thread_type = W24ThreadType.WHITWORTH

    whitworth_size: Decimal

    tolerance_class: Optional[str] = None


class W24ThreadFeature(W24FeatureModel):
    """Characterization of a Thread Feature

    Attributes:
        gender: Gender (male or female) of the thread feature.
            This is determined by checking whether the thread is
            located on the outer contour of the part or inside the part.
            When the outer contour is unavailable (e.g., in detail drawings),
            the gender is set to None.

        length: Length of the slug corresponding
            to the thread. Please keep in mind that a thus can
            contain multiple threads with different lengths. This
            is considered in the threads themselves.
            NOTE: we are using the word length here to differentiate
            it from the thread depth which describes the difference
            between the major and minor radii.

        threads (List[W24Thread]): List of Threads that are positioned
            on the ThreadFeatures. This is a list to support multi-threads

        NOTE: Tapers are currently not considered

        NOTE: Future implementations might also consider the inclination
        in two angles relative to the front view.
    """

    gender: Optional[W24Gender]

    length: Optional[Decimal]

    threads: List[W24Thread]

    @validator('threads', pre=True, allow_reuse=True)
    def ask_list_validator(
        cls,
        raw: List[Dict[str, Any]]
    ) -> List[W24Thread]:
        """ Validator to de-serialize the asks. The de-serialization
        is based on the ask_type attribute of the object. Pydantic
        does not support this out-of-the box

        Args:
            raw (Dict[str, Any]): Raw json of the asks list

        Returns:
            List[W24AskUnion]: List of deserialized Asks
        """
        return [_deserialize_thread_type(t) for t in raw]


def _deserialize_thread_type(
    thread_type: str
) -> Type[W24Thread]:
    """ Get the Ask Class from the ask type

    Args:
        ask_type (str): Ask type in question

    Raises:
        ValueError: Raised if ask type is unknown

    Returns:
        str: Name of the AskObject
    """
    class_ = {
        "ISO_METRIC": W24ThreadISOMetric,
        "WHITWORTH": W24ThreadWhitworth,
        "UTS": W24ThreadUTS,
    }.get(thread_type, None)

    if class_ is None:
        raise ValueError(f"Unknown Ask Type '{thread_type}'")

    return class_



from enum import Enum
from typing import Any, Dict, List, Optional

from pydantic import BaseModel, validator

from werk24.models.color import W24Color

from .feature import W24FeatureModel
from .general_tolerances import W24GeneralTolerances
from .language import W24Language
from .material import W24Material
from .weight import W24Weight


class W24TitleBlockItem(W24FeatureModel):
    """ Per-Language caption or value

    Attributes:

        language: Language in accordance with the ISO/639-2B standards

        text: Text of the identification
    """

    language: Optional[W24Language]

    text: str


class W24CaptionValuePair(BaseModel):
    """ Caption-Value pair for that were found on the Title Block.

    Attributes:

        blurb: Caption-Value pair for human consumption

        captions: List of captions in different languages.
            This will only return the languages that were detected and
            NOT translate the captions into languages that are not present
            on the drawing. This behavior might however change in the
            future.
    """
    blurb: str

    captions: List[W24TitleBlockItem]

    values: List[W24TitleBlockItem]


class W24FileExtensionType(str, Enum):
    """ 
    Enum of the extension types.

    For example, pdf and idw extensions will be mapped to
    DRAWING, while step and stl extensions will be mapped
    to MODEL.
    """
    DRAWING = "DRAWING"
    MODEL = "MODEL"
    UNKNOWN = "UNKNOWN"


class W24FilePathType(str, Enum):
    """ 
    Enum of the file path types, indicating whether a
    POSIX (unix) or WINDOWS path is used. When only a filename
    is indicated, the value will be UNKNOWN
    """
    POSIX = "POSIX"
    WINDOWS = "WINDOWS"
    UNKNOWN = "UNKNOWN"


class W24Filename(W24FeatureModel):
    """ 
    Object describing all the information that we can
    deduce from a filename that was found on the TitleBlock

    Attributes:

        blurb: Filename and Path as it was found on the drawing.
            Example: /path/to/drawing.pdf

        filename: Filename without the prefix.
            Example drawing.pdf

        extension: Extension of the filename.
            Examples: .pdf, .tar.gz

        extension_type: filetype indicated by the extension. PDF, ...
            extensions are mapped to DRAWING, while STEP, ...
            extensions are mapped to MODEL

        path_type: WINDOWS or POSIX (unix) path types if we have
            found an absolute path. UNKNOWN if we only read
            a filename without prefix.
    """
    blurb: str

    filename: str

    extension: str

    extension_type: W24FileExtensionType

    path_type: W24FilePathType


class W24TitleBlock(BaseModel):
    """ 
    Information that could be extracted from the Title Block.

    Attributes:

        designation: Designation of the Sheet on the Title Block

        drawing_id: Main Identification Number of the Drawing

        part_ids: List of Part IDs that are located on the drawing.
            Keep in mind that the drawing might define multiple
            variants that each specify their own part id

        reference_ids: List of additional reference IDs
            detected on the Drawing

        general_tolerances: General Tolerances quoted on the TitleBlock

        material: Material which is quoted on the TitleBlock

        weight: Weight as read from the TitleBlock. NOTE: this is not
            cross-checked with the material and volume of the part,
            but provided as it was read on the TitleBlock.

        filename_drawing: Filename of the drawing if it is explicitly
            indicated on the title block

        colors: List of colors detected on the TitleBlock or in the
            canvas notes.
    """

    designation: Optional[W24CaptionValuePair]

    drawing_id: Optional[W24CaptionValuePair]

    part_ids: List[W24CaptionValuePair] = []

    reference_ids: List[W24CaptionValuePair]

    general_tolerances: Optional[W24GeneralTolerances]

    material: Optional[W24Material]

    weight: Optional[W24Weight]

    filename_drawing: Optional[W24Filename] = None

    colors: List[W24Color] = []

    @validator('designation', pre=True)
    def designation_validator(
        cls,
        raw: Dict[str, Any]
    ) -> Optional[W24CaptionValuePair]:
        """ 
        Workaround to deal with the transition period
        while we move from the single-value to the multi-value
        pairs.

        This code can be removed after we complete the
        transition.

        Args:

            raw (Dict[str, Any]): Unparsed value returned
                from the API

        Returns:

            W24CaptionValuePair: Parse value-caption pair
        """
        return cls._parse_caption_value_pair(raw)

    @validator('drawing_id', pre=True)
    def drawing_id_validator(
        cls,
        raw: Dict[str, Any]
    ) -> Optional[W24CaptionValuePair]:
        """ 
        Workaround to deal with the transition period
        while we move from the single-value to the multi-value
        pairs.

        This code can be removed after we complete the
        transition.

        Args:

            raw (Dict[str, Any]): Unparsed value returned
                from the API

        Returns:

            W24CaptionValuePair: Parse value-caption pair
        """
        return cls._parse_caption_value_pair(raw)

    @validator('reference_ids', pre=True)
    def reference_ids_validator(
        cls,
        raw: List[Dict[str, Any]]
    ) -> List[W24CaptionValuePair]:
        """ 
        Workaround to deal with the transition period
        while we move from the single-value to the multi-value
        pairs.

        This code can be removed after we complete the
        transition.

        Args:

            raw (List[Dict[str, Any]]): Unparsed value returned
                from the API

        Returns:

            List[W24CaptionValuePair]: Parse value-caption pair
        """
        result = [
            cls._parse_caption_value_pair(e)
            for e in raw
        ]
        return [r for r in result if r is not None]

    @staticmethod
    def _parse_caption_value_pair(
        raw: Optional[Dict[str, Any]]
    ) -> Optional[W24CaptionValuePair]:
        """ 
        Workaround to deal with the transition period
        while we move from the single-value to the multi-value
        pairs.

        This code can be removed after we complete the
        transition.

        Args:

            raw (Dict[str, Any]): Unparsed value returned
                from the API

        Returns:

            W24CaptionValuePair: Parse value-caption pair
        """
        if isinstance(raw, W24CaptionValuePair):
            return raw

        if raw is None:
            return None

        if 'value' in raw.keys():
            raw['values'] = [{
                'language': None,
                'text': raw.get('value')
            }]
            del raw['value']

        return W24CaptionValuePair.parse_obj(raw)



""" Command Line Interface for W24 Techread
"""
import argparse
import io
import json
import logging
from collections import namedtuple
from typing import Any, Dict, List, Optional

from dotenv import load_dotenv
from pydantic import UUID4, ValidationError

from werk24.cli import utils
from werk24.exceptions import RequestTooLargeException
from werk24.models.ask import (
    W24AskCanvasThumbnail,
    W24AskPageThumbnail,
    W24AskSectionalThumbnail,
    W24AskSheetAnonymization,
    W24AskSheetThumbnail,
    W24AskTitleBlock,
    W24AskVariantCAD,
    W24AskVariantExternalDimensions,
    W24AskVariantGDTs,
    W24AskVariantMeasures,
    W24AskVariantRadii,
    W24AskVariantRoughnesses,
    W24AskVariantThreadElements,
)
from werk24.models.techread import (
    W24TechreadException,
    W24TechreadMessageSubtypeError,
    W24TechreadMessageSubtypeProgress,
    W24TechreadMessageType,
)
from werk24.techread_client import Hook

# load the environment variables
load_dotenv(".werk24")

# set the log level to info for the test setting
# We recommend using logging.WARNING for production
logging.basicConfig(
    level=logging.INFO,
    format='%(asctime)s.%(msecs)03d %(levelname)s %(module)s - %(funcName)s: %(message)s',  # noqa
    datefmt='%Y-%m-%d %H:%M:%S')

# get the local logger
logger = logging.getLogger(__name__)  # pylint:disable = invalid-name


# make the configuration of what hooks we want to handle and how
HookConfig = namedtuple("HookConfig", "arg ask function")
hook_config = [
    HookConfig(
        'ask_page_thumbnail',
        W24AskPageThumbnail,
        lambda m: _show_image("Page Thumbnail", m.payload_bytes)),
    HookConfig(
        'ask_sheet_thumbnail',
        W24AskSheetThumbnail,
        lambda m: _show_image("Sheet Thumbnail", m.payload_bytes)),
    HookConfig(
        'ask_sheet_anonymization',
        W24AskSheetAnonymization,
        lambda m: _show_image("Sheet Anonymization", m.payload_bytes)
    ),
    HookConfig(
        'ask_canvas_thumbnail',
        W24AskCanvasThumbnail,
        lambda m: _show_image("Canvas Thumbnail", m.payload_bytes)),
    HookConfig(
        'ask_sectional_thumbnail',
        W24AskSectionalThumbnail,
        lambda m: _show_image("Sectional Thumbnail", m.payload_bytes)),
    # HookConfig(
    #     'ask_variant_angles',
    #     W24AskVariantAngles,
    #     lambda m: _print_payload("Ask Variant Angles", m.payload_dict)),
    HookConfig(
        'ask_variant_external_dimensions',
        W24AskVariantExternalDimensions,
        lambda m: _print_payload("Ask Variant Ext. Dimensions", m.payload_dict)),
    HookConfig(
        'ask_variant_gdts',
        W24AskVariantGDTs,
        lambda m: _print_payload("Ask Variant GDTs", m.payload_dict)),
    HookConfig(
        'ask_variant_measures',
        W24AskVariantMeasures,
        lambda m: _print_payload("Ask Variant Measures", m.payload_dict)),
    HookConfig(
        'ask_variant_radii',
        W24AskVariantRadii,
        lambda m: _print_payload("Ask Variant Radii", m.payload_dict)),
    HookConfig(
        'ask_variant_roughnesses',
        W24AskVariantRoughnesses,
        lambda m: _print_payload("Ask Variant Roughnesses", m.payload_dict)),
    HookConfig(
        'ask_variant_cad',
        W24AskVariantCAD,
        lambda m: _store_variant_cad(
            m.payload_dict,
            m.payload_bytes,
            m.exceptions)),
    HookConfig(
        'ask_titleblock',
        W24AskTitleBlock,
        lambda m: _print_payload("Ask TitleBlock", m.payload_dict)
    ),
    HookConfig(
        'ask_variant_thread_elements',
        W24AskVariantThreadElements,
        lambda m: _print_payload("Ask Variant Thread Elements", m.payload_dict)
    ),
]


def _store_variant_cad(
    payload_dict: Dict[str, Any],
    payload_bytes: Optional[bytes],
    exceptions: List[W24TechreadException]
) -> None:
    """ Store the CAD file the current directory

    Args:
        payload_dict (Dict[str, Any]): Payload Dictionary
        payload_bytes (bytes): CAD that we received as response
    """
    logger.info(f"Ask Variant CAD\n{payload_dict}")

    # print potential exceptions
    if exceptions:
        _log_exceptions(exceptions)

    # store the payload
    if payload_bytes is not None:
        _store_variant_cad_payload(payload_dict, payload_bytes)


def _log_exceptions(
    exceptions: List[W24TechreadException]
) -> None:
    """ Log the encountered exceptions as warnings

    Args:
        exceptions (List[W24TechreadException]): List of
            encountered exceptions. Can be an empty list.
    """
    for cur_exception in exceptions:
        logger.warn(cur_exception)


def _store_variant_cad_payload(
    payload_dict: Dict[str, Any],
    payload_bytes: bytes
) -> None:
    """ Store the CAD file with the knowledge that
    the payload exists

    Args:
        payload_dict (Dict[str, Any]): Payload Dictionary
        payload_bytes (bytes): CAD response
    """
    # make the filename
    variant_id = payload_dict.get('variant_id')
    filename = f"./w24_ask_variant_cad_{variant_id}.dxf"

    # and write the content
    with open(filename, "wb+") as file_handle:
        file_handle.write(payload_bytes)

    # tell the user
    logger.info(f"CAD response stored in {filename}")


def _get_drawing(
    file_path: str
) -> Optional[bytes]:
    """ Get the bytes of the file that shall be
    read.

    Args:
        file_path (str): path to the input file

    Returns:
        bytes: content of the fiel
    """
    try:
        with open(file_path, "rb") as filehandle:
            return filehandle.read()
    except FileNotFoundError:
        print(f"File {file_path} not found")
        return None


def _print_payload(
    log_text: str,
    payload_dict: Dict[str, Any]
) -> None:
    """ Display the payload dictionary in a format that
    is easy for humans to read

    Args:
        log_text (str): Headline info
        payload_dict (Dict[str, Any]): Payload dictionary
    """
    print(log_text)
    payload_json = json.dumps(payload_dict, indent=4)
    print(payload_json)


def _show_image(
        log_text: str,
        image_bytes: bytes
) -> None:
    """ Display a debug image. The function relies on
    the PIL functionality of Image.show(). The
    actual behavior will thus be different across platforms

    Args:
        log_text (str): Log Text that we want to write back
            to the command line interface
        image_bytes (bytes): byte representation of the image
            that is to be displayed
    """

    # Import the PIL at runtime to make sure that
    # you can still use the client without the need
    # to install Pillow
    try:
        from PIL import Image  # pylint: disable=import-outside-toplevel
    except ImportError:
        logger.warning(
            "Viewing image-like responses requires the installation"
            " of the PIL package: pip install pillow. Image skipped.")
        return

    # otherwise print the text
    logger.info(log_text)

    # and show the image
    image = Image.open(io.BytesIO(image_bytes))
    try:
        image.show(title=log_text)
    except BaseException:
        logger.warning(
            "Image cannot be displayed. "
            "Please check your local security settings.")
        return


async def main(
    args: argparse.Namespace
) -> None:
    """
    Run the Techread command with the CLI arguments

    Args:
        args(argparse.Namespace): CLI args generated by
        argparse
    """

    # interpret the sub-account as UUID4 and stop if that
    # does not conform to the correct pattern.
    try:
        if args.sub_account is not None:
            sub_account = UUID4(args.sub_account)
        else:
            sub_account = None

    except ValueError:
        logger.error("Sub-Account ID does not conform to UUID4-pattern")
        return

    # make the hooks from the arguments
    hooks = _make_hooks_from_args(args)

    # get the client instance and handle
    # potential errors
    client = utils.make_client()

    async with client as session:

        # get the drawing
        drawing_bytes = _get_drawing(args.input_file)
        if drawing_bytes is None:
            return

        # and make the request
        try:
            await session.read_drawing_with_hooks(
                drawing_bytes,
                hooks,
                sub_account=sub_account)

        except RequestTooLargeException:
            message = "Request was too large to be processed. " \
                + "Please check the documentation for current limits."
            logger.error(message)


def _make_hooks_from_args(
        args: argparse.Namespace
) -> List[Hook]:

    # add a general hook to deal with internal errors
    hook_error_internal = Hook(
        message_type=W24TechreadMessageType.ERROR,
        message_subtype=W24TechreadMessageSubtypeError.INTERNAL,
        function=lambda msg:
        logger.error("Internal Error %s", msg.payload_dict))

    # add a general hook to deal with PROGRESS messages
    hook_progress_started = Hook(
        message_type=W24TechreadMessageType.PROGRESS,
        message_subtype=W24TechreadMessageSubtypeProgress.STARTED,
        function=lambda msg: logger.info("Techread process started"))

    # tell the api what asks you are interested in,
    # and define what to do when you receive the result
    hooks = [
        Hook(ask=c.ask(), function=c.function)
        for c in hook_config
        if getattr(args, c.arg)
    ] + [
        hook_error_internal,
        hook_progress_started
    ]

    return hooks



"""Definition of all W24Ask types that are understood by the Werk24 API.
"""
from typing import Tuple
from enum import Enum
from typing import Any, Dict, List, Optional, Type, Union

from pydantic import UUID4, BaseModel, HttpUrl

from werk24.models.note import W24Note
from werk24.models.part_family import W24PartFamilyCharacterization

from .angle import W24Angle
from .file_format import W24FileFormatThumbnail, W24FileFormatVariantCAD
from .gdt import W24GDT
from .general_tolerances import W24GeneralTolerances
from .geometric_shape import W24GeometricShapeCuboid, W24GeometricShapeCylinder
from .leader import W24Leader
from .material import W24Material
from .measure import W24Measure
from .radius import W24Radius
from .revision_table import W24RevisionTable
from .roughness import W24Roughness
from .thread_element import W24ThreadElement


class W24AskType(str, Enum):
    """List of all Ask Type supported by the current
    API version. This list will grow with future releases.

    """

    CANVAS_THUMBNAIL = "CANVAS_THUMBNAIL"
    """Thumbnail of the canvas (i.e., the part of the
    sheet that contains the geometry)
    """

    NOTES = "NOTES"
    """Notes of the Sectional and the Canvas
    """

    PAGE_THUMBNAIL = "PAGE_THUMBNAIL"
    """Thumbnail of the overall page - rotated and with
    surrounding white space removed
    """

    PART_FAMILY_CHARACTERIZATION = "PART_FAMILY_CHARACTERIZATION"
    """Ask that triggers a post processor corresponding to the
    part family
    """
    PRODUCT_PMI_EXTRACT = "PRODUCT_PMI_EXTRACT"
    """Ask for the PMI Extract Product
    """

    REVISION_TABLE = "REVISION_TABLE"
    """Ask for the Revision Table
    """

    SECTIONAL_THUMBNAIL = "SECTIONAL_THUMBNAIL"
    """Thumbnail of a sectional on the canvas.
    Here the sectional describes both cuts and perspectives
    """

    SHEET_THUMBNAIL = "SHEET_THUMBNAIL"
    """Thumbnail of the sheet (i.e., the part of the
    page that is described by the surrounding frame)
    """

    SHEET_ANONYMIZATION = "SHEET_ANONYMIZATION"
    """Thumbnail of the sheet with all references to
    the original author removed.
    """

    TITLE_BLOCK = "TITLE_BLOCK"
    """Ask for all information that is available on the
    title block
    """

    TRAIN = "TRAIN"
    """Supplying the request for training only without
    expecting a response.
    """

    VARIANT_ANGLES = "VARIANT_ANGLES"
    """Requests the all Angles on the variant
    """

    VARIANT_CAD = "VARIANT_CAD"
    """Requests the generation of a CAD file
    """

    VARIANT_GDTS = "VARIANT_GDTS"
    """List of Geometric Dimensions and Tolerances detected
    on the Sectionals associated with the variant
    """

    VARIANT_LEADERS = "VARIANT_LEADERS"
    """List of Leaders that were detected on the Sectional
    """

    VARIANT_MATERIAL = "VARIANT_MATERIAL"
    """Material that was detected on the data fields of the
    drawing or within a variant table
    """

    VARIANT_MEASURES = "VARIANT_MEASURES"
    """List of Measures that were found on the Sectionals
    associated with the variant
    """

    VARIANT_RADII = "VARIANT_RADII"
    """List of all Radii that were found on teh Sectionals
    associated with the variant
    """

    VARIANT_ROUGHNESSES = "VARIANT_ROUGHNESSES"
    """List of Roughnesses that were found on the Sectionals
    associated with the variant
    """

    VARIANT_EXTERNAL_DIMENSIONS = "VARIANT_EXTERNAL_DIMENSIONS"
    """Ask for the external dimensions
    """

    VARIANT_THREAD_ELEMENTS = "VARIANT_THREAD_ELEMENTS"
    """Ask for the thread elements of the variant
    """

    VARIANT_TOLERANCE_ELEMENTS = "VARIANT_TOLERANCE_ELEMENTS"
    """Ask for the tolerance elements of the variant
    """

    INTERNAL_SCREENING = "INTERNAL_SCREENING"
    """Ask for internal screening - not available through public API
    """


class W24Ask(BaseModel):
    """Base model from which all Asks inherit

    Attributes:

        ask_type: Type of the requested Ask. Used
            for de-serialization.

        is_training: Flag that indicates that your request
            is a pure training request and that you are not
            expecting to obtain a response.

    !!! note
        When you send a request with the attribute is_training=True
        you are directly improving the quality of our Machine
        Learning Models with regards to your domain. These requests
        are not charged, but they also do not generate a response.
        The connection is immediately closed after you submit the
        request; our system them processes the drawing when the
        system load is low.
    """
    ask_type: W24AskType

    is_training: bool = False


class W24AskThumbnail(W24Ask):
    """Base model for features that request a thumbnail.

    Attributes:

        file_format: File format in which you wish to obtain
            the result. Currently only JPEG is supported.

    !!! note
        At this stage, the API will return a high-resolution
        gray-level image. Future releases might allow you to
        request color images or to set a resolution limit.
        If this is a priority to you, please let us know.
    """
    file_format: W24FileFormatThumbnail = W24FileFormatThumbnail.JPEG


class W24AskPageThumbnail(W24AskThumbnail):
    """Requests a thumbnail for each page in the document;
    rotated, and with the surrounding white-space removed.

    !!! note
        We preprocess the page so that it is always white-on-black,
        even when the Technical Drawing that you submitted was
        black-on-white.
    """
    ask_type = W24AskType.PAGE_THUMBNAIL


class W24AskSheetThumbnail(W24AskThumbnail):
    """Requests a thumbnail of each sheet on each page in
    the document. The sheet will only contain the pixels within
    the main frame that surrounds the canvas and header fields.

    !!! note
        We preprocess the sheet so that it is always white-on-black,
        even when the Technical Drawing that you submitted was
        black-on-white.
    """
    ask_type = W24AskType.SHEET_THUMBNAIL


class W24AskSheetAnonymization(W24AskThumbnail):
    """Requests an ANONYMIZED thumbnail of each sheet on each page
    in the document. The sheet will only contain the pixels within
    the main frame that surrounds the canvas and header fields.

    !!! note
        We preprocess the sheet so that it is always white-on-black,
        even when the Technical Drawing that you submitted was
        black-on-white.

    Attributes:

        replacement_logo (bytes): SVG version of the logo that
            shall be used to replace the logo on the drawing.

        identification_snippets (List[str]): List of string values
            that might identify the authoring company. The software
            will actively look for such text segments in the data
            and replace them with white pixels.

        redact_employee_names (bool): Redact the employee names
            when they are indicated by a caption such as
            "Drawn by"

        redact_cage_code (bool): Redact the CAGE Code (
            Commercial and Government Entity Code) typically
            used for US-government suppliers

        fill_color (Tuple[int,int,int]): Color that shall be used
            to fill the pixels that are to be redacted. Currently
            only grayscale is supported.

    """
    ask_type = W24AskType.SHEET_ANONYMIZATION

    replacement_logo_url: Optional[HttpUrl] = None

    identification_snippets: List[str] = []

    redact_employee_names: bool = False

    redact_cage_code: bool = False

    fill_color: Tuple[int, int, int] = (255, 255, 255)


class W24AskPartFamilyCharacterization(W24Ask):
    """Triggers a post-processor that turns the raw data into a
    dictionary of attributes characteristic for the part family.

    E.g. a Screw could have the following attributes:
    * Drive Type
    * Head diameter
    * Head length
    * Thread
    * Thread length

    The definition of a part family can be done in a Part Family
    Template and is implemented by Werk24 on the backend - where
    it has access to more granular data.

    Attributes:

        part_family_id (UUID4): Unique part family identifier
            that we will provide to you after the part family
            post processor is implemented.
    """
    ask_type = W24AskType.PART_FAMILY_CHARACTERIZATION

    part_family_id: UUID4


class W24AskPartFamilyCharacterizationResponse(BaseModel):
    """Response object corresponding to a PartFamilyCharacterization request.


    Attributes:

        page_id (UUID4): Id of the page that specified the part_family

        sheet_id (UUID4):Id of the sheet
    """
    page_id: UUID4
    sheet_id: UUID4
    part_family_characterizations: List[W24PartFamilyCharacterization]


class W24AskCanvasThumbnail(W24AskThumbnail):
    """Requests a thumbnail of each canvas in each sheet.
    The canvas describes the "drawing area" of the sheet.

    !!! note
        We preprocess the canvas so that it is always white-on-black,
        even when the Technical Drawing that you submitted was
        black-on-white.

    Attributes:

        remove_canvas_notes__dangerous (bool): Remove the canvas
            notes from the canvas thumbnail.

            !!! DANGEROUS: chances are that you are removing
            !!! important information. Run a risk analysis
            !!! before using this attribute
    """
    ask_type = W24AskType.CANVAS_THUMBNAIL

    remove_canvas_notes__dangerous: bool = False


class W24AskSectionalThumbnail(W24AskThumbnail):
    """The W24AskPlaneThumbnail requests a thumbnail
    of each sectional on each sheet in the document.

    !!! note
        We preprocess the sectional so that it is always white-on-black,
        even when the Technical Drawing that you submitted was
        black-on-white.
    """
    ask_type = W24AskType.SECTIONAL_THUMBNAIL


class W24AskVariantAngles(W24Ask):
    """With this Ask you are requesting the list of all
    measures that were detected on all sectionals of a
    variant.
    """
    ask_type = W24AskType.VARIANT_ANGLES


class W24AskVariantAnglesResponse(BaseModel):
    """ResponseType associated with the
    W24AskVariantAngles.

    Attributes:

        variant_id: Unique ID of the variant detected on
            the Technical Drawing. Refer to the documentation
            on Variants for details.

        sectional_id: Unique ID of the sectional on which the
            Angle was detected. This allows you to associate
            the Angle to the SectionalThumbnail (should you
            have requested it).

        angles: List of Angles that were found for the Variant
            on the Sectional.
    """
    variant_id: UUID4
    sectional_id: UUID4
    angles: List[W24Angle]


class W24AskVariantRoughnesses(W24Ask):
    """With this Ask you are requesting the list of all
    roughnesses (surface symbols) that were detected for
    the variant.
    """

    ask_type = W24AskType.VARIANT_ROUGHNESSES


class W24AskVariantRoughnessesResponse(BaseModel):
    """Response object corresponding to the
    W24AskVariantRoughnesses ask.

    Attributes:

        variant_id: Unique ID of the variant detected on
            the Technical Drawing. Refer to the documentation
            on Variants for details.

        sectional_id: Unique ID of the sectional on which the
            Measure was detected. This allows you to associate
            the Measure to the SectionalThumbnail (should you
            have requested it).

        roughnesses: List of Roughnesses that were found for the
            Variant on the Sectional.
    """
    variant_id: UUID4
    sectional_id: UUID4
    roughnesses: List[W24Roughness]


class W24AskVariantRadii(W24Ask):
    """With this Ask you are requesting the list of all
    radii that were detected for the variant.
    """
    ask_type = W24AskType.VARIANT_RADII


class W24AskVariantRadiiResponse(BaseModel):
    """Response object corresponding to the
    W24AskVariantRadii ask.

    Attributes:

        variant_id: Unique ID of the variant detected on
            the Technical Drawing. Refer to the documentation
            on Variants for details.

        sectional_id: Unique ID of the sectional on which the
            Measure was detected. This allows you to associate
            the Measure to the SectionalThumbnail (should you
            have requested it).

        radii: List of Radii that were found for the
            Variant on the Sectional.
    """
    variant_id: UUID4
    sectional_id: UUID4
    radii: List[W24Radius]


class W24AskVariantMeasures(W24Ask):
    """With this Ask you are requesting the complete
    list of all measures that were detected for the
    variant.

    Attributes:

        confidence_min: Werk24 calculates internal
            confidence scores for each measure. Depending
            on your use-case you might want to consider or
            discard low-confidence results. This parameter
            allows you to filter the results. The resulting
            W24Measure objects also contain a confidence score
            that allows you to filter even further.
    """
    ask_type = W24AskType.VARIANT_MEASURES

    confidence_min: float = 0.2


class W24AskVariantMeasuresResponse(BaseModel):
    """Response object corresponding to the
    W24AskVariantMeasures ask.

    Attributes:

        variant_id: Unique ID of the variant detected on
            the Technical Drawing. Refer to the documentation
            on Variants for details.

        sectional_id: Unique ID of the sectional on which the
            Measure was detected. This allows you to associate
            the Measure to the SectionalThumbnail (should you
            have requested it).

        measures: List of Measures that were found for the
            Variant on the Sectional.

    !!! note
        Be aware that requesting the measures will
        yield one responds for each variant and sectional
    """
    variant_id: UUID4
    sectional_id: UUID4
    measures: List[W24Measure]


class W24AskVariantLeaders(W24Ask):
    """With this Ask you are requesting the complete
    list of all leaders that were detected on the
    variant

    !!! danger
        This feature is currently in invitation-only beta
        and will only be answered if this features has been
        activated for your account. Otherwise the request
        will be ignored.
    """
    ask_type = W24AskType.VARIANT_LEADERS


class W24AskVariantLeadersResponse(BaseModel):
    """Response object corresponding to the
    W24AskVariantLeaders.


    Attributes:

        variant_id: Unique ID of the variant detected on
            the Technical Drawing. Refer to the documentation
            on Variants for details.

        sectional_id: Unique ID of the sectional on which the
            Leader was detected. This allows you to associate
            the Leader to the SectionalThumbnail (should you
            have requested it).

        leaders: List of Leaders that were found for the
            Variant on the Sectional.
    """
    variant_id: UUID4
    sectional_id: UUID4
    leaders: List[W24Leader]


class W24AskVariantMaterial(W24Ask):
    """This ask requests the material of the individual variant.
    It will now only consider the material that is listed on the
    TitleBlock, but also consider material information that is
    found as text on the Canvas.

    Attributes:

        material_hint: If already know the material that "should"
            be indicated on the drawing, you are welcome to submit
            it here as hint. This will allow us to improve the
            accuracy of our algorithms. Feel free to use any material
            designation that is convenient.

    !!! danger
        This feature is currently in invitation-only beta
        and will only be answered if this features has been
        activated for your account. Otherwise the request
        will be ignored.
    """
    ask_type = W24AskType.VARIANT_MATERIAL

    material_hint: Optional[str] = None


class W24AskTitleBlock(W24Ask):
    """This ask requests all information that
    we can obtain from the title block
    """
    ask_type = W24AskType.TITLE_BLOCK


class W24AskRevisionTable(W24Ask):
    """With this Ask you are requesting the list of all
    revision tables in the document
    """
    ask_type = W24AskType.REVISION_TABLE


class W24AskRevisionTableResponse(BaseModel):
    """Response object corresponding toi the
    W24AskRevisionTable

    Attributes:

        revision_table: RevisionTable object with all
            the content that was extracted from the
            drawing
    """
    revision_table: W24RevisionTable


class W24AskVariantGDTs(W24Ask):
    """This Ask requests the list of all
    Geometric Dimensions and Tolerances
    that were detected for the Variant.
    """
    ask_type = W24AskType.VARIANT_GDTS


class W24AskVariantGDTsResponse(BaseModel):
    """Response object corresponding to the
    W24AskVariantGDTs.

    Attributes:

        variant_id: Unique ID of the variant detected on
            the Technical Drawing. Refer to the documentation
            on Variants for details.

        sectional_id: Unique ID of the sectional on which the
            GD&T was detected. This allows you to associate
            the Leader to the SectionalThumbnail (should you
            have requested it).

        gdts: List of GD&Ts that were found for the
            Variant on the Sectional.
    """
    variant_id: UUID4
    sectional_id: UUID4
    gdts: List[W24GDT]


class W24AskTrain(W24Ask):
    """If you submit this Ask, we will use your request
    to train and improve our models. It does not trigger a response.

    !!! danger
        This is deprecated. Please use the attribute is_training=True
        instead.
    """
    ask_type = W24AskType.TRAIN


class W24AskVariantCAD(W24Ask):
    """By sending this Ask, you are requesting
    an associated CAD model

    Attributes:

        output_format: Output format in which to generate
            the CAD file.

    !!! note
        This Ask will currently return a DXF approximation
        of a flat part and can only be used for 2 dimensional
        applications (e.g. sheet metal).
    """
    ask_type = W24AskType.VARIANT_CAD

    output_format: W24FileFormatVariantCAD = W24FileFormatVariantCAD.DXF


class W24AskVariantCADResponse(BaseModel):
    """Response object corresponding to the W24AskVariantCad.

    Attributes:

        variant_id: Unique ID of the variant detected on
            the Technical Drawing. Refer to the documentation
            on Variants for details.

        num_sectionals: Number of sectionals that were detected
            on the Variant. This allows you to better classify
            the type of file you are dealing with.

        num_angles: Number of  angles that were detected on the Variant.
            This allows you to better classify the type of file you are
            dealing with.



    !!!! note
        Be aware that requesting the measures will yield one response
        for each variant. Refer to the Variant documentation for
        details

    !!! note
        The cad file will be returned as part of the payload_bytes
        and needs to be accessed directly. If you are using your own
        client implementation, you need to obtain the information from
        the payload_url (and submit the correct token).


    !!! danger
        The attributes num_sectionals and num_angles will be deprecated
        soon in favor of a part classifier. Please reach out to us
        before using these two attributes.

    """
    variant_id: UUID4

    num_sectionals: int = 0

    num_angles: int = 0


class W24AskVariantExternalDimensions(W24Ask):
    """Ask object to request the external dimensions of each
    variant on the Document.
    """

    ask_type = W24AskType.VARIANT_EXTERNAL_DIMENSIONS


class W24AskVariantExternalDimensionsResponse(BaseModel):
    """Response object corresponding to the W24AskVariantExternalDimensions

    Attributes:

        variant_id (UUID4): Unique ID of the variant detected on the
            Technical Drawing. Refer to the documentation on Variants
            for details.

        enclosing_cuboid (W24GeometricShapeCuboid): Cuboid that encloses
            the complete geometric shape of the part

        enclosing_cylinder (W24GeometricShapeCuboid): Cylinder that encloses
            the complete geometric shape of the part

        confidence (float): Confidence that the reading is correct.

        !!! the cuboid and cylinder are not necessarily aligned. So might
        !!! receive different depths of the cuboid and the cylinder.
    """
    variant_id: UUID4
    enclosing_cuboid: Optional[W24GeometricShapeCuboid]
    enclosing_cylinder: Optional[W24GeometricShapeCylinder]
    confidence: float


class W24AskProductPMIExtract(W24Ask):
    """Ask object to request the PMIExtract Product.
    """
    ask_type = W24AskType.PRODUCT_PMI_EXTRACT


class W24AskProductPMIExtractResponse(BaseModel):
    """Response object corresponding to the W24AskProduct PMIExtract

    Attributes:

        variant_id (UUID4): Unique ID of the variant detected on the
            Technical Drawing. Refer to the documentation on Variants
            for details.

        material (Optional[W24Material]): Material that was detected on
            the technical drawing. If no material was detected,
            this is set to None.

        general_tolerances (Optional[W24GeneralTolerances]): General tolerances
            detected on the drawing. This will automatically translate the
            general tolerances detected on the canvas notes to an ISO-2768
            class. None if no general tolerances are detected.

        measures (List[W24Measure]): List of the available measures on
            the drawing. Note: in the PMIExtract, the position will not be
            returned.

        gdts (List[W24GDT]): List of the detected GD&Ts. Note: in the
            PMIExtract, the position will not be returned.

        radii (List[W24Radius]): List of the detected Radii. Note: in the
            PMIExtract, the position will not be returned.

        roughnesses (List[W24Roughness]): List of the detected
            roughnesses. Note: in the PMIExtract, the position will not
            be returned.
    """
    variant_id: UUID4
    material: Optional[W24Material]
    general_tolerances: Optional[W24GeneralTolerances]
    measures: List[W24Measure]
    gdts: List[W24GDT]
    radii: List[W24Radius]
    roughnesses: List[W24Roughness]


class W24AskVariantThreadElements(W24Ask):
    """Ask object to obtain the thread elements
    """
    ask_type = W24AskType.VARIANT_THREAD_ELEMENTS


class W24AskVariantThreadElementsResponse(BaseModel):
    """Response object corresponding to the W24AskVariantThreadElements

    Attributes:

        variant_id (UUID4): Unique ID of the variant detected on the
            Technical Drawing. Refer to the documentation on Variants
            for details.

    """
    variant_id: UUID4
    sectional_id: UUID4
    thread_elements: List[W24ThreadElement]


class W24AskNotes(W24Ask):
    """Ask all the notes on the Canvas and the sectionals
    """
    ask_type = W24AskType.NOTES


class W24AskNotesResponse(BaseModel):
    """Response to the W24AskNotes

    Attributes:

        notes(List[W24Ask]): List of the notes that were
            identified.
    """
    notes: List[W24Note]


class W24AskInternalScreening(W24Ask):
    """Internal Ask to trigger the file screening.

    NOTE: not available on the public API.
    """
    ask_type = W24AskType.INTERNAL_SCREENING


# class W24AskVariantToleranceElements(W24Ask):
#     """Ask object to obtain the tolerance elements
#     """
#     ask_type = W24AskType.VARIANT_TOLERANCE_ELEMENTS

# class W24AskVariantToleranceElementsResponse(BaseModel):
#     """Response object corresponding to the W24AskVariantThreadElements

#     Attributes:

#         variant_id (UUID4): Unique ID of the variant detected on the
#             Technical Drawing. Refer to the documentation on Variants
#             for details.

#     """
#     variant_id: UUID4
#     thread_elements: List[W24ToleranceElement]

W24AskUnion = Union[
    W24AskCanvasThumbnail,
    W24AskNotes,
    W24AskPageThumbnail,
    W24AskPartFamilyCharacterization,
    W24AskProductPMIExtract,
    W24AskRevisionTable,
    W24AskSectionalThumbnail,
    W24AskSheetAnonymization,
    W24AskSheetThumbnail,
    W24AskTitleBlock,
    W24AskTrain,
    W24AskVariantExternalDimensions,
    W24AskVariantCAD,
    W24AskVariantGDTs,
    W24AskVariantLeaders,
    W24AskVariantMaterial,
    W24AskVariantMeasures,
    W24AskVariantRadii,
    W24AskVariantRoughnesses,
    W24AskVariantThreadElements,
    # W24AskVariantToleranceElements
]
"""Union of all W24Asks to ensure proper de-serialization """


def deserialize_ask(
    raw: Union[Dict[str, Any], W24Ask],
) -> W24Ask:
    """Deserialize a specific ask in its raw form

    Args:
        raw (Dict[str, Any]): Raw Ask as it arrives from the
            json deserializer

    Returns:
        W24AskUnion: Corresponding ask type
    """
    if isinstance(raw, dict):
        ask_type = _deserialize_ask_type(raw.get('ask_type', ''))
        return ask_type.parse_obj(raw)

    if isinstance(raw, W24Ask):
        return raw

    raise ValueError(f"Unsupported value type '{type(raw)}'")


def _deserialize_ask_type(
    ask_type: str
) -> Type[W24Ask]:
    """Get the Ask Class from the ask type

    Args:

        ask_type (str): Ask type in question

    Raises:

        ValueError: Raised if ask type is unknown

    Returns:

        str: Name of the AskObject
    """
    class_ = {
        "CANVAS_THUMBNAIL": W24AskCanvasThumbnail,
        "INTERNAL_SCREENING": W24AskInternalScreening,
        "NOTES": W24AskNotes,
        "PAGE_THUMBNAIL": W24AskPageThumbnail,
        "PART_FAMILY_CHARACTERIZATION": W24AskPartFamilyCharacterization,
        "PRODUCT_PMI_EXTRACT": W24AskProductPMIExtract,
        "REVISION_TABLE": W24AskRevisionTable,
        "SECTIONAL_THUMBNAIL": W24AskSectionalThumbnail,
        "SHEET_ANONYMIZATION": W24AskSheetAnonymization,
        "SHEET_THUMBNAIL": W24AskSheetThumbnail,
        "TITLE_BLOCK": W24AskTitleBlock,
        "TRAIN": W24AskTrain,
        "VARIANT_EXTERNAL_DIMENSIONS": W24AskVariantExternalDimensions,
        "VARIANT_GDTS": W24AskVariantGDTs,
        "VARIANT_LEADERS": W24AskVariantLeaders,
        "VARIANT_MATERIAL": W24AskVariantMaterial,
        "VARIANT_MEASURES": W24AskVariantMeasures,
        "VARIANT_RADII": W24AskVariantRadii,
        "VARIANT_ROUGHNESSES": W24AskVariantRoughnesses,
        "VARIANT_CAD": W24AskVariantCAD,
        "VARIANT_THREAD_ELEMENTS": W24AskVariantThreadElements,
        # "VARIANT_TOLERANCE_ELEMENTS":W24AskVariantToleranceElements,
    }.get(ask_type, None)

    if class_ is None:
        raise ValueError(f"Unknown Ask Type '{ask_type}'")

    return class_




""" Defintion of all the W24Measure class its support structures
"""

from enum import Enum
from typing import Any, Dict, List, Optional

from pydantic import UUID4, BaseModel, validator

from .chamfer import W24Chamfer
from .depth import W24Depth
from .feature import W24FeatureModel
from .size import (W24Size, W24SizeTolerance, W24SizeToleranceGeneral,
                   parse_tolerance)
from .test_dimension import W24TestDimension
from .thread import W24Thread
from .unit import W24UnitLength


class W24MeasureWarningType(str, Enum):
    """ 
    List of all warnings that can be associated with
    the returned measures.
    """

    UNCONVENTIONAL_TOLERANCE_ORDER = "UNCONVENTIONAL_TOLERANCE_ORDER"
    """ 
    The UNCONVENTIONAL_TOLERANCE_ORDER warning is raised
    when the first-mentioned tolerance is lower than the second-mentioned.

    EXAMPLE: 3 -0.1/+0.1 (rather than 3 +0.1/-0.1)
    """

    UNPROPORTIONAL = "UNPROPORTIONAL"
    """ 
    The UNPROPORTIONAL warnings is set when the size indicated
    on the MeasureLabel is unproportional to the distance between
    the Measure's end-points.
    """


class W24MeasureWarning(BaseModel):
    """ 
    Warnings are issued when something about the label
    or measure is not conforming with convention
    """
    warning_type: W24MeasureWarningType


class W24MeasureWarningUnconvetionalToleranceOrder(W24MeasureWarning):
    """ 
    The UNCONVENTIONAL_TOLERANCE_ORDER warning is raised
    when the first-mentioned tolerance is lower than the second-mentioned.

    EXAMPLE: 3 -0.1/+0.1 (rather than 3 +0.1/-0.1)
    """
    warning_type = W24MeasureWarningType.UNCONVENTIONAL_TOLERANCE_ORDER


class W24MeasureWarningUnproportional(W24MeasureWarning):
    """ 
    The UNPROPORTIONAL warnings is set when the size indicated
    on the MeasureLabel is unproportional to the distance between
    the Measure's end-points.

    !!! note Three things can cause this:

    1. The Measure is truly unpropotional and carries an associated
        indicator.

    2. The Measure is unporportional, but not indicated as such.
        Typically seen on drawings that "were modified on the fly"

    3. The size of the MeasureLabel was misread by the Werk24
        backend.

    Due to the lack of training data, we can currently not
    differentiate between case 1 and case 2. Once the training
    data suffices, this warning will be extended by boolean
    field 'indicated'.

    In addition, it is theoretically possible to estimate
    the likelihood of case 2 and case 3. If this becomes a
    true concern in your application, please reach out to us
    """
    warning_type = W24MeasureWarningType.UNPROPORTIONAL


class W24MeasureLabel(BaseModel):
    """ Measure Label

    Attributes:

        blurb: String representation of the item for human consumption

        quantity: Quantity for spacings, e.g., (2x). Currently only the
            indicated measure will be detected and returned. Future
            implementation could detect the spacings and return individual
            measures for the respective spacings. When we implement this
            feature, we will add an attribute to the ASK, which allows you
            to control the behavior.

        size: Size of measure as referred to in the drawing.
            Sizes are always associated with units! This becomes important
            to remember when you are dealing with mixed-unit drawings (e.g.,
            an adapter bolt that has an ISO thread and an UTC tapped hole).
            To avoid any loss of precision, we return the size in the unit
            that was indicated on the drawing.

        size_tolerance: Tolerance details. Default: General tolerances.
            By default we are referring to the general tolerances of the drawing.
            When you are also requesting the TitleBlock, we are attaching the
            applicable General Tolerances (assuming they were understood)
            to the measure. This will consider the nominal size of the
            measure.
            If the W24MeasureLabel describes a "Theoretically Exact Measure",
            i.e, the label is surrounded by a box, like: "[15]", the size_tolerance
            refers to a W24SizeToleranceTheoreticallyExact object (and is NOT None)

        unit: Length unit of the size. This information is only available
            when you request the TitleBlock as well; as only the title block
            will allow us to determine the geographic origin of the drawing.

        thread: Optional thread details.
            The thread details describe the complete thread description
            and follow the respective standards. In consequence, the thread
            diameter of an ISO-thread will be indicated in millimeter,
            while the thread diameter of an UTS thread will be in inch.

        chamfer: Optional Chamfer

        depth: Depth of the drilling or thread. Uses the same dimensions

    """

    blurb: str

    quantity: int = 1

    size: W24Size

    size_tolerance: W24SizeTolerance = W24SizeToleranceGeneral()

    unit: Optional[W24UnitLength] = None

    thread: Optional[W24Thread] = None

    chamfer: Optional[W24Chamfer] = None

    depth: Optional[W24Depth] = None

    test_dimension: Optional[W24TestDimension] = None

    @validator('size_tolerance', pre=True)
    def asks_validator(  # NOQA
        cls,
        raw: Dict[str, Any]
    ) -> Optional[W24SizeTolerance]:
        """ Pydantic does not automatically return the correct
        W24SizeTolerance object. This function looks at the toleration_type
        attribute and returns the correct W24SizeTolerance subclass

        Args:

            size_tolerance_raw (Dict[str, str]): Raw Dictionary of
                the size tolerance

        Returns:
            W24SizeTolerance: Correctly deserialized Size Tolerance
        """
        return parse_tolerance(raw)


class W24Measure(W24FeatureModel):
    """ Measure object

    Attributes:

        measure_id: Unique UUID4 identifier

        label: Measure Label

        warnings: List of Warnings that are associated with the
            measure. See W24MeasureWarning for details.

        confidence:  Werk24 calculates an internal confidence score for
            each measure. Depending on your use-case, you might want
            to consider or discard low-confidence measures. This
            value allows you to do so. The value ranges from
            0.0 to 1.0

    """
    measure_id: Optional[UUID4] = None

    label: W24MeasureLabel

    warnings: List[W24MeasureWarning] = []

    confidence: float = 0.0



from enum import Enum


class W24Language(str, Enum):
    """ Enum of supported languages following
    the naming convention of ISO/639-2B
    """

    DEU = "DEU"
    """ German """

    DUT = "DUT"
    """ Dutch """

    ENG = "ENG"
    """ English """

    FRA = "FRA"
    """ French """

    ITA = "ITA"
    """ Italian """

    SPA = "SPA"
    """ Spanish """

    UNKNOWN = "__UNK__"
    """ Unknown """



# from PyQt5 import Qsci


# class W24GuiJsonEditor(Qsci.QsciScintilla):
#     """ Small JSON Editor with highlighter functionality
#     to make it a bit easier for the user to read the
#     JSON responses that come from the API
#     """

#     def __init__(self, parent: int = None) -> None:
#         super().__init__(parent)
#         self.setLexer(Qsci.QsciLexerJSON(self))



"""Data Models describing part families.

Author: Jochen Mattes
"""
from typing import Any, Dict

from pydantic import UUID4, BaseModel


class W24PartFamilyCharacterization(BaseModel):
    """Instance of a part family described by the File.

    Part families can be defined by the Werk24 partner.
    This allows us the flexibility to consider the exact
    situation of the customer.

    Take two companies that both deal with screws, but use
    different characteristic values to describe the part
    family. With this flexibility, we can simply specify
    a postprocessor for each customer and are able to
    consider their very special situation.

    Attributes:
        part_family_id (UUID4): Unique Identifier of the part
            family. You need this ID to trigger the correct
            postprocessor. Note that this is account-specific,
            so a post-processors need to be activated for your
            account, but are available to all of your tenants.
            If you want to show them selectively to your
            users, you are responsible for checking that the
            postprocessor can only be used by authorized
            tenants / users.

        part_family_name (str): A more human way of referring
            to the part family. This identifier is chosen by
            Werk24 according to an internal naming convention.

        variant_id (UUID4): Unique identifier of the variant.
            Note that a single drawing can specify multiple
            part family instances.

        characteristic_values (Dict[str, Any]): Characteristic
            values that define the part family. When exported to
            CSV, the keys correspond the column names, and the
            values to the cell values.
    """
    part_family_id: UUID4
    part_family_name: str
    variant_id: UUID4
    characteristic_values: Dict[str, Any]



from typing import Optional

from .feature import W24FeatureModel
from .unit import W24UnitWeight, W24UnitLength


class W24Weight(W24FeatureModel):
    """ Weight of a Part - for example as indicated
    on the Title Block.

    NOTE: this can also be a relative weight
    (e.g. kg per meter for profiles)

    Attributes:
        blurb: Weight in a human-readable format.

        value: Value of the Weight

        weight_unit: Unit of the Weight (e.g., kg)

        length_unit: Unit of the Length for relative
            weights (e.g., METER for meter weights of profiles)s
    """
    blurb: str

    value: float

    weight_unit: W24UnitWeight
    length_unit: Optional[W24UnitLength]



from datetime import date
from typing import List, Optional

from pydantic import BaseModel

from .date import W24Date


class W24GridSquare(BaseModel):
    """ Grid Square

    Attributes:
        column (str): Column of the Grid Square

        row (str): Row of the Grid Square

    """
    column: str

    row: str


class W24RevisionTableRow(BaseModel):
    """ Row of a revision table

    Attributes:
        serial (Optional[str]): Serial number,
            also referred to as Index or Position of
            the row in the revision table. The serial number
            is typicall used to identify the revision on
            each individual sheet (e.g., A, B, C or 01, 02, 03)

        ecn_number (Optional[str]): Identifier of the Engineering
            Change Note to uniquely identify the change across multiple
            documents. This is typically a long number and sometimes
            identifies a revision process that affect multiple parts.

        description (str): Description of the change

        revision_date (Optional[date]): Date of the revision

        grid_squares: (List[W24GridSquare]): List of the grid squares
            that are affected by the revision.
    """
    serial: Optional[str]

    ecn_number: Optional[str]

    description: str

    revision_date: Optional[W24Date]

    grid_squares: List[W24GridSquare]


class W24RevisionTable(BaseModel):
    """ Revision Table object that contains all the
    rows

    Attributes:
        rows (List[W24RevisionTableRow]): List of all
            rows in the revision table
    """
    rows: List[W24RevisionTableRow]




from enum import Enum


class W24UnitLength(str, Enum):
    """ Enum of the supported length units
    """
    METER = "METER"
    DECIMETER = "DECIMETER"
    CENTIMETER = "CENTIMETER"
    MILLIMETER = "MILLIMETER"
    FOOT = "FOOT"
    INCH = "INCH"
    MICRO_INCH = "MICRO_INCH"


class W24UnitAngle(str, Enum):
    """ Enum of all the supported angle units

    NOTE: currently we are only supporting degrees.
    If you come across a drawing that uses GON, let
    us know, we love the concept!
    """
    DEGREE = "DEGREE"


class W24UnitWeight(str, Enum):
    """ Enum of all the supported weights

    NOTE: This also includes relative weights
    such as kilogram per meter

    """

    # Absolute weights
    GRAM = "GRAM"
    KILOGRAM = "KILOGRAM"
    POUND = "POUND"
    OUNCE = "OUNCE"


class W24UnitSystem(str, Enum):
    """ Unit System that is used for a certain feature
    """
    METRIC = "METRIC"
    IMPERIAL = "IMPERIAL"



""" Data Model for Canvas and Sectional Notes.

Author: Jochen Mattes
"""
from enum import Enum

from werk24.models.feature import W24FeatureModel


class W24NoteType(str, Enum):
    """Differentiation between Canvas and Sectional Notes.

    Canvas Notes are sometimes referred to as "General Notes",
    they are associated with the canvas in general.

    Sectional Notes are associated with a specific sectional,
    but not as part of a leader or call-out. (e.g., "VIEW A")

    Sectional Call Outs are linked to the sectional through
    a leader symbol (e.g., "MARK HERE")

    NOTE: please note that we are only returning the information
    that is not understood by the system through another way.
    Over time, number of sectional call outs will thus gradually
    reduce.
    """
    CANVAS_NOTE = "CANVAS NOTE"
    SECTIONAL_NOTE = "SECTIONAL NOTE"
    SECTIONAL_CALLOUT = "SECTIONAL_CALLOUT"


class W24Note(W24FeatureModel):
    """Notes object for Sectional and Canvas Notes.

    Attributes:
        type (W24NoteType): Type of the note to differentiate
            between notes that are associated with the complete
            sectional and the canvas

        text (str): Raw text that was read.
    """
    type: W24NoteType
    text: str



from decimal import Decimal
from itertools import permutations
from typing import Tuple

from pydantic import BaseModel


class W24GeometricShapeCuboid(BaseModel):
    """ Geometric Shape of a cuboid

    Attributes:
        width (Decimal): Width of the cuboid

        height (Decimal): Height of the cuboid

        depth (Decimal): Depth of the cuboid
    """
    width: Decimal
    height: Decimal
    depth: Decimal

    def to_tuple(self) -> Tuple[Decimal, Decimal, Decimal]:
        """ Get a `width`x`depth`x`height` tuple

        Returns:
            Tuple[Decimal,Decimal, Decimal]: Width, height, depth
                tuple of the cuboid
        """
        return (self.width, self.height, self.depth)

    def encloses(
        self,
        other: 'W24GeometricShapeCuboid',
        allow_width_depth_rotation: bool = True,
    ) -> bool:
        """ Does this cuboid enclose `other` cuboid?

        Check whether this cuboid encloses the `other` cuboid.
        The `allow_width_depth_rotation` parameter allows you
        to specify whether the edges of the two cuboids
        need to be parallel.

        Args:
            other (W24GeometricShapeCuboid): Other cuboid
            allow_width_depth_rotation (bool, optional): Allow
                to rotate the `other` cuboid around this
                cuboid's width/depth plane. Especially useful
                for additive manufacturing applications.
                Defaults to True.

        Raises:
            RuntimeError: Raised when receiving a object other
                than a cuboid

        Returns:
            bool: True if `self` encloses `other`.
        """

        # ensure that we are actually dealing with a
        # cuboid. NOTE: in the future we might want to
        # support cylinders as well
        if not isinstance(other, W24GeometricShapeCuboid):
            raise RuntimeError("Invalid datatype for `other`")

        # check whether the part fits into the machine
        # in a fully parallel way.
        if self._encloses__parallel(other):
            return True

        # if rotation on the width-depth plane is allowed,
        # check that as well
        if allow_width_depth_rotation \
                and self._encloses__width_depth_rotation(other):
            return True

        return False

    def _encloses__parallel(
        self,
        other: 'W24GeometricShapeCuboidShape'
    ) -> bool:
        """ Check whether the other cuboid fits into this
        cuboid without rotating the other cuboid on any
        axis. The `other` cuboid can only be rotated by
        fully 90 degrees (i.e, we are allowed to swap
        the width and height)

        Args:
            other (W24GeometricShapeCuboidShape): Other cuboid
                that should be fit into the `self` cuboid

        Returns:
            bool: True if the `other` cuboid fits into the
                `self` cuboid while preserving the parallelism
                between the two cuboids
        """
        machine = sorted(self.to_tuple())
        part = sorted(other.to_tuple())
        return all(m >= p for m, p in zip(machine, part))

    def _encloses__width_depth_rotation(
        self,
        other: 'W24GeometricShapeCuboidShape'
    ) -> bool:

        # without loss of generality: consider the machine width
        # to be the larger of the width/depth dimensions
        machine_width = float(max(self.width, self.depth))
        machine_depth = float(min(self.width, self.depth))
        diagonal_length = (machine_width**2+machine_depth**2)**0.5
        if machine_depth == 0:
            return False
        aspect_ratio = machine_depth/machine_width

        # go through all possible rotations of the part
        # PERFORMANCE: this will generate at most 6 iterations
        for width, height, depth in set(permutations(other.to_tuple(), 3)):

            # check whether the width fits into the diagonal
            if width > diagonal_length:
                continue

            # check the height first (cheaper to check)
            if height > self.height:
                continue

            # check whether there is still enough space left
            # between the end of the depth-edge of the part
            # and the machine depth
            max_depth = (diagonal_length-float(width))*aspect_ratio
            if depth > max_depth:
                continue

            # if all criteria are met, we can accept
            return True

        return False


class W24GeometricShapeCylinder(BaseModel):
    """ Geometric Shape of a cylinder

    Attributes:
        diameter (Decimal): Diameter of the cylinder

        depths (Decimal): Depth of the cylinder
    """
    diameter: Decimal
    depth: Decimal


W24GeometricShapeCuboid.update_forward_refs()



from decimal import Decimal

from pydantic import BaseModel

from .unit import W24UnitAngle


class W24Chamfer(BaseModel):
    """ Chamfer in degree.

    Attributes:
        blurb: String representation for human consumption

        angle: Chamfer angle in degrees

        unit: Angle Unit. Currently only degrees are
            supported.

    !!! note
        The chamfers can be tolerated.
        Future implementations will take this into account.

    """

    blurb: str

    angle: Decimal

    unit: W24UnitAngle = W24UnitAngle.DEGREE



from .unit import W24UnitLength
from decimal import Decimal
from enum import Enum
from typing import List, Optional

from pydantic import UUID4, BaseModel

from .feature import W24FeatureModel
from .unit import W24UnitSystem


class W24RoughnessStandard(str, Enum):
    """ Most standards that define the surface roughness use
    very similar symbols. However, the position of the fields
    varies.

    The standards listed here are understood and supported
    by the API.

    NOTE: the ISO 1302 standard exists in four different versions,
    these releases substantially modified the position and
    structure of the roughness symbols.

    NOTE: this list is not exhaustive, many countries have specified
    their own standards over the years. Do not hesitate to reach
    out if you wish us to implement another standard.

    """
    ISO_1302_1978 = "ISO 1302:1978"
    ISO_1302_1992 = "ISO 1302:1992"
    ISO_1302_2002 = "ISO 1302:2002"
    ISO_21920_1_2021 = "ISO 21920-1:2021"
    ASME_Y14_36_1978 = "ASME Y14.36-1978"
    ASME_Y14_36M_1996 = "ASME Y14.36M-1996"
    ASME_Y14_36_2018 = "ASME Y14.36-2018"


class W24RoughnessMaterialRemovalType(str, Enum):
    """ Most standard allow the designer to specify
    whether material removal is required or prohibited.

    By default both options are allowed.
    """
    UNSPECIFIED = "UNSPECIFIED"
    PROHIBITED = "PROHIBITED"
    REQUIRED = "REQUIRED"


class W24RoughnessDirectionOfLay(str, Enum):
    """ The lay of the roughness limits the
    manufacturing process and is sometimes
    required for the application.
    """
    PARALLEL = "="
    PERPENDICULAR = "⟂"
    CROSS = "X"
    MULTIDIRECTIONAL = "M"
    CIRCULAR = "C"
    RADIAL = "R"
    PROTUBERANT = "P"


class W24RoughnessAcceptanceCriterion(str, Enum):
    """ The designer can specify whether to apply
    the 16%-rule, the maximum- or medium- rule
    when deciding whether a surface complies with
    the specifications.
    """
    SIXTEEN_PERCENT = "16%"
    MAXIMUM = "max"
    MEAN = "mean"


class W24RoughnessConditionType(str, Enum):
    """ Roughnesses can specify, the
    upper limit, the lower limit and the
    average.
    """
    UPPER = "U"
    LOWER = "L"
    AVERAGE = "A"


class W24RoughnessFilterType(str, Enum):
    """ When measuring the roughness, different
    filter types will result in different results.
    To reduce this source of ambiguity, the
    designer can specify the filter to be used.

    NOTE: this list is not exhaustive. If you
    need us to support another filter method,
    please reach out.
    """
    GAUSSIAN = "G"
    ROBUST_GAUSSIAN = "RG"
    SPLINE = "S"
    TWO_RC = "\"2RC\""


class W24RoughnessGrade(BaseModel):
    """ ISO Roughness Grade

    Attributes:
        blurb: String representation for human consumption

        grade: Roughness Grade in the range N01 to N12.
            Valid values are N01, N0, N1, N2, N3, N4, N5,
            N6, N7, N8, N9, N10, N11, N12, N13, N14

    """
    blurb: str

    grade: str


class W24RoughnessParameter(str, Enum):
    """ Roughness Parameter that is specified.

    NOTE: this list is not exhaustive, but
    covers the most frequently used parameters.
    """

    # ASME
    AA = "AA"
    CLA = "CLA"
    PVA = "PVA"
    RMS = "RMS"

    # PROFILE
    PA = "Pa"
    PC = "PC"
    PKU = "Pku"
    PP = "Pp"
    PQ = "Pq"
    PSK = "Psk"
    PT = "Pt"
    PV = "Pv"
    PY = "Py"
    PZ = "Pz"

    # ROUGHNESS
    RA_DIN = "Ra(DIN)"
    RA_ISO = "Ra(ISO)"
    RC = "Rc"
    RKU = "Rku"
    RP = "Rp"
    RQ = "Rq"
    RSK = "Rsk"
    RT = "Rt"
    RV = "Rv"
    RY = "Ry"
    RZ = "Rz"

    # WAVINESS
    WA = "Wa"
    WC = "Wc"
    WKU = "Wku"
    WP = "Wp"
    WQ = "Wq"
    WSK = "Wsk"
    WT = "Wt"
    WV = "Wv"
    WY = "Wy"
    WZ = "Wz"

    # N-GRADE
    N = "N"


class W24RoughnessEvaluationLengthType(str, Enum):
    LENGTH = "LENGTH"
    LAMBDA_C_MULTIPLE = "LAMBDA_C_MULTIPLE"


class W24RoughnessEvaluationLength(BaseModel):
    """ For sophisticated application, the sample needs
    to be taken over a longer distances. This is specified the
    either the sampling length in millimeter (ISO 3012:1974,
    ISO 3012:1978, ISO 3012:1992) or as multiple of the main lambda
    (ISO 3012:2002 and ISO 3012:2021).


    Attributes:
        evaluation_length_type: Sampling length type that is specified
            by the applicable standard

        length: evaluation length in the specified units.

        length_unit: Millimeter for both the ISO and ASME standards

        lambda_c_multiple: multiple of the main cutoff lambda_c

    """
    evaluation_length_type: W24RoughnessEvaluationLengthType

    length: Optional[Decimal]
    length_unit: Optional[W24UnitLength]

    lambda_c_multiple: Optional[Decimal]


class W24RoughnessCondition(BaseModel):
    """ Roughness Label

    Attributes:
        blurb: String representation of the Roughness for human
            consumption.

        condition_type: Roughness Side that is specified (upper or lower).
            We automatically determine this value depending on
            the standard and the position of the parameter

        filter_type: Filter Type to use when measuring the surface.
            When not specified on the drawing, we set the default
            value according to the applicable standard.

        lambda_s: Filter bandwidth microroughness cutoff lambda_s.
            When not specified, we determine the default value based
            on the standard and the value_limit.

        lambda_c: Filter bandwidth main cutoff lambda_c.
            When not specified, we determine the default value based
            on the standard and the value_limit.

        parameter: Parameter to be measured


        characteristic: Method of converting the 2D-measurement into
            a single number

        evaluation_length: For sophisticated application, the roughness needs
            to be evaluated over a longer distances. This is specified the
            either the sampling length in millimeter (ISO 3012:1974,
            ISO 3012:1978, ISO 3012:1992) or as multiple of the main lambda
            (ISO 3012:2002 and ISO 3012:2021).

        acceptance_criterion: Designers can specify what rule to
            apply when deciding whether a measured tolerance complies
            with the requirements

        value_limit: Specification limit

        roughness_grade: ISO roughness grade that the roughness specifications
            correspond to. IMPORTANT: this allows you to deal with
            all the standards in a simple way.

    """
    blurb: str

    condition_type: Optional[W24RoughnessConditionType]

    filter_type: Optional[W24RoughnessFilterType]

    lambda_s: Optional[Decimal]

    lambda_c: Optional[Decimal]

    parameter: Optional[W24RoughnessParameter]

    evaluation_length: Optional[W24RoughnessEvaluationLength]

    acceptance_criterion: Optional[W24RoughnessAcceptanceCriterion]

    value: Optional[Decimal]

    roughness_grade: Optional[W24RoughnessGrade]


class W24ManufacturingMethod(BaseModel):
    """ Manufacturing method specified on the drawing.
    Currently only the blurb is available.

    Attributes:
        blurb: Manufacturing method as specified on the
            drawing
    """
    blurb: str


class W24RoughnessWaviness(BaseModel):
    """ Roughness Waviness as defined in the ASME-Y14-36

    Attributes:
        waviness_height: Waviness height in inch
        waviness_width: Waviness width in inch
    """
    waviness_height: Decimal
    waviness_width: Decimal


class W24RoughnessLabel(BaseModel):
    """ Roughness Label

    Attributes:
        blurb: String for human consumption

        blurb_html: HTML for human consumption. The complex structure of the
            symbol makes this necessary.

        standard : Standard in which the Roughness Symbol is written
            NOTE: there are multiple standards that are in use

        machining_allowance: Machining Allowance in millimeter of micro-inch.
            See the `unit` attribute.

        material_removal_type: The MaterialRemovalType specifies whether material
            must / must not be removed.

        applies_all_around: Boolean value that specifies whether the roughness
            applies to all surfaces around

        direction_of_lay: The lay of the roughness limits the manufacturing
            process and is sometimes required for the application.

        manufacturing_method: Method by which to achieve the roughness

        conditions: Roughness conditions

        unit_system: Unit system that is used for the roughness.
            We are not using the explicit units as the different
            attributes have different units (e.g, micro and nano meters)

        waviness: Waviness defined in ASME standards

    """
    blurb: str

    blurb_html: str

    standard: W24RoughnessStandard

    machining_allowance: Optional[Decimal]

    material_removal_type: W24RoughnessMaterialRemovalType

    applies_all_around: bool

    direction_of_lay: Optional[W24RoughnessDirectionOfLay]

    manufacturing_method: W24ManufacturingMethod

    conditions: List[W24RoughnessCondition]

    unit_system: W24UnitSystem

    waviness: Optional[W24RoughnessWaviness]


class W24Roughness(W24FeatureModel):
    """ Roughness object

    Attributes:
        roughness_id: Unique UUID4 identifier

        label: RoughnessLabel
    """
    roughness_id: UUID4

    label: W24RoughnessLabel
